# -*- coding: utf-8 -*-
"""ELM Web Scraping - Base class for search engine search"""
import os
import random
import asyncio
import logging
import requests
from urllib.parse import quote
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager

from serpapi.serp_api_client import SerpApiClient
from rebrowser_playwright.async_api import async_playwright
from playwright_stealth import StealthConfig

from elm.web.utilities import PWKwargs, clean_search_query, pw_page


logger = logging.getLogger(__name__)


class SearchEngineLinkSearch(ABC):
    """Abstract class to retrieve links for a query using a search engine"""

    _SE_NAME = "<unknown se>"

    async def results(self, *queries, num_results=10):
        """Retrieve links for the first `num_results` of each query

        This function executes a search for each input query and
        returns a list of links corresponding to the top `num_results`.

        Parameters
        ----------
        *queries : str
            One or more queries to search for.
        num_results : int, optional
            Maximum number of top results to retrieve for each query.
            Note that this value can never exceed the number of results
            per page (typically 10). If you pass in a larger value, it
            will be reduced to the number of results per page. There is
            also no guarantee that the search query will return this
            many results - the actual number of results returned is
            determined by the number of results on a page (excluding
            ads). You can, however, use this input to limit the number
            of results returned. By default, ``10``.

        Returns
        -------
        list
            List equal to the length of the input queries, where each
            entry is another list containing no more than `num_results`
            links.
        """
        queries = map(clean_search_query, queries)
        return await self._get_links(queries, num_results)

    async def _get_links(self, queries, num_results):
        """Get links for multiple queries"""
        outer_task_name = asyncio.current_task().get_name()

        searches = [
            asyncio.create_task(
                self._skip_exc_search(query, num_results=num_results),
                name=outer_task_name,
            )
            for query in queries
        ]
        logger.trace("Kicking off %s search for %d queries",
                     self._SE_NAME, len(searches))
        results = await asyncio.gather(*searches)
        logger.trace("Got results for link search:\n%r", results)
        return results

    async def _skip_exc_search(self, query, num_results=10):
        """Perform search while ignoring errors"""
        try:
            return await self._search(query, num_results=num_results)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            msg = "Could not complete search for query=%r\nGot error type: %r"
            logger.exception(msg, query, type(e))
            return []

    async def _move_mouse(self, page):
        """Simulate mouse movement"""
        logger.trace("Moving mouse")
        vp = page.viewport_size or {}
        logger.trace("VP: %r", vp)
        width, height = vp.get('width', 300), vp.get('height', 300)
        min_width, max_width = int(width * 0.1), int(width * 0.9)
        min_height, max_height = int(height * 0.1), int(height * 0.9)
        logger.trace("Moving mouse to random position within: %d-%d, %d-%d",
                     min_width, max_width, min_height, max_height)

        await page.mouse.move(random.randint(min_width, max_width),
                              random.randint(min_height, max_height),
                              steps=10)
        await asyncio.sleep(random.uniform(1.5, 3.5))

    async def _move_and_click(self, page, input_el):
        """Move mouse to an element and click on it"""
        box = await input_el.bounding_box()
        if box is None:
            return await input_el.click()

        x = box["x"] + int(box["width"] / random.uniform(1.1, 10))
        y = box["y"] + int(box["height"] / random.uniform(1.1, 10))

        await page.mouse.move(x, y, steps=random.randint(5, 30))
        await asyncio.sleep(random.uniform(0.2, 0.6))
        return await page.mouse.click(x, y)

    @abstractmethod
    async def _search(self, query, num_results=10):
        """Search web for links related to a query"""
        raise NotImplementedError


class PlaywrightSearchEngineLinkSearch(SearchEngineLinkSearch):
    """Abstract class to search the web using playwright and a search engine"""

    MAX_RESULTS_CONSIDERED_PER_PAGE = 10
    """Number of results considered per search engine page"""

    PAGE_LOAD_TIMEOUT = 60_000
    """Default page load timeout value in milliseconds"""

    _SC = StealthConfig(navigator_user_agent=False)

    def __init__(self, use_homepage=True, use_scrapling_stealth=False,
                 **launch_kwargs):
        """

        Parameters
        ----------
        use_homepage : bool, default=True
            If ``True``, the browser will be navigated to the search
            engine homepage and the query will be input into the search
            bar. If ``False``, the query will be embedded in the URL
            and the browser will navigate directly to the filled-out
            URL. By default, ``False``.
        use_scrapling_stealth : bool, default=False
            Option to use scrapling stealth scripts instead of
            tf-playwright-stealth. If set to ``True``, the `_SC` class
            attribute will be ignored. By default, ``False``.
        **launch_kwargs
            Keyword arguments to be passed to
            `playwright.chromium.launch`. For example, you can pass
            ``headless=False, slow_mo=50`` for a visualization of the
            search.
        """
        self.use_homepage = use_homepage
        self.use_scrapling_stealth = use_scrapling_stealth
        self.launch_kwargs = PWKwargs.launch_kwargs()
        self.launch_kwargs.update(launch_kwargs)
        self._browser = None

    async def _load_browser(self, pw_instance):
        """Launch a chromium instance and load a page"""
        self._browser = await pw_instance.chromium.launch(**self.launch_kwargs)

    async def _close_browser(self):
        """Close browser instance and reset internal attributes"""
        logger.trace("Closing browser...")
        if self._browser is None:
            return

        await self._browser.close()
        self._browser = None

    @asynccontextmanager
    async def _browser_page(self):
        """Get page to use for search"""
        page_kwargs = {"browser": self._browser, "stealth_config": self._SC,
                       "ignore_https_errors": True,  # no sensitive inputs
                       "timeout": self.PAGE_LOAD_TIMEOUT,
                       "use_scrapling_stealth": self.use_scrapling_stealth}

        async with pw_page(**page_kwargs) as page:
            yield page

    async def _search(self, query, num_results=10):
        """Search web for links related to a query"""
        logger.debug("Searching %s: %r", self._SE_NAME, query)
        num_results = min(num_results, self.MAX_RESULTS_CONSIDERED_PER_PAGE)

        async with self._browser_page() as page:
            if self.use_homepage:
                logger.trace("Navigating to %s homepage", self._SE_NAME)
                await _navigate_to_se_url(page, se_url=self._SE_URL,
                                          timeout=self.PAGE_LOAD_TIMEOUT)
                logger.trace("Performing %s search for query: %r",
                             self._SE_NAME, query)
                await self._perform_homepage_search(page, query)
            else:
                url = self._SE_QUERY_URL.format(quote(query))
                logger.trace("Submitting URL: %r", url)
                await _navigate_to_se_url(page, se_url=url,
                                          timeout=self.PAGE_LOAD_TIMEOUT)
            logger.trace("Extracting links for query: %r", query)
            return await self._extract_links(page, num_results, query)

    async def _get_links(self, queries, num_results):
        """Get links for multiple queries"""
        outer_task_name = asyncio.current_task().get_name()
        async with async_playwright() as pw_instance:
            await self._load_browser(pw_instance)
            searches = [
                asyncio.create_task(
                    self._skip_exc_search(query, num_results=num_results),
                    name=outer_task_name,
                )
                for query in queries
            ]
            logger.trace("Kicking off %s search for %d queries",
                         self._SE_NAME, len(searches))
            results = await asyncio.gather(*searches)
            logger.trace("Got results for link search:\n%r", results)
            await self._close_browser()
        return results

    async def _extract_links(self, page, num_results, query):
        """Extract links for top `num_results` on page"""
        await page.wait_for_load_state("networkidle",
                                       timeout=self.PAGE_LOAD_TIMEOUT)
        await page.wait_for_selector(self._SE_SR_TAG)
        locator = page.locator(self._SE_SR_TAG)
        count = await locator.count()
        links = []

        for i in range(count):
            element = locator.nth(i)
            try:
                link = await element.get_attribute("href")
                if link is not None:
                    links.append(link)
            except Exception:
                logger.exception("Skipped extracting link %d for query %r",
                                 i, query)

            if len(links) >= num_results:
                break

        return links

    @property
    @abstractmethod
    def _SE_URL(self):
        """str: Search engine URL"""
        raise NotImplementedError

    @property
    @abstractmethod
    def _SE_SR_TAG(self):
        """str: Search engine search results tag"""
        raise NotImplementedError

    @property
    @abstractmethod
    def _SE_QUERY_URL(self):
        """str: Search engine query URL template"""
        raise NotImplementedError

    @abstractmethod
    async def _perform_homepage_search(self, page, search_query):
        """Search query using search engine homepage"""
        raise NotImplementedError


class APISearchEngineLinkSearch(SearchEngineLinkSearch):
    """Abstract class to search the web using a search engine API"""

    API_KEY_VAR = None
    """Name of environment variable that should contain the API key"""

    def __init__(self, api_key=None):
        """

        Parameters
        ----------
        api_key : str, optional
            API key for search engine. If ``None``, will look up the API
            key using the :obj:`API_KEY_VAR` environment variable.
            By default, ``None``.
        """
        self.api_key = api_key or os.environ.get(self.API_KEY_VAR or "")


class PatchedSerpApiClient(SerpApiClient):
    """SerpApiClient patched to allow bypassing of SSL verification"""

    def __init__(self, params_dict, engine=None, timeout=60000, verify=True):
        super().__init__(params_dict=params_dict, engine=engine,
                         timeout=timeout)
        self.verify = verify

    def get_response(self, path='/search'):
        """Get search response

        Parameters
        ----------
        path : str, default='/search'
            API path to use for the search.

        Returns
        -------
            Response object provided by ``requests.get``.
        """
        url = None
        try:
            url, parameter = self.construct_url(path)
            response = requests.get(url, parameter, timeout=self.timeout,
                                    verify=self.verify)
            return response
        except requests.HTTPError as e:
            logger.error("fail: " + url)
            logger.error(e, e.response.status_code)
            raise e


async def _navigate_to_se_url(page, se_url, timeout=90_000):
    """Navigate to search engine url"""
    await page.goto(se_url)
    logger.trace("Waiting for load")
    await page.wait_for_load_state("networkidle", timeout=timeout)



# -*- coding: utf-8 -*-
"""ELM Ordinance full processing logic"""
import time
import json
import asyncio
import logging
from datetime import datetime, timedelta
from pathlib import Path
from functools import partial

import openai
import pandas as pd
from langchain.text_splitter import RecursiveCharacterTextSplitter

from elm import ApiBase
from elm.utilities import validate_azure_api_params
from elm.ords.download import download_county_ordinance
from elm.ords.extraction import (
    extract_ordinance_text_with_ngram_validation,
    extract_ordinance_values,
)
from elm.ords.services.usage import UsageTracker
from elm.ords.services.openai import OpenAIService, usage_from_response
from elm.ords.services.provider import RunningAsyncServices
from elm.ords.services.threaded import (
    TempFileCache,
    FileMover,
    CleanedFileWriter,
    OrdDBFileWriter,
    UsageUpdater,
)
from elm.ords.services.cpu import PDFLoader, read_pdf_doc, read_pdf_doc_ocr
from elm.ords.utilities import (
    RTS_SEPARATORS,
    load_all_county_info,
    load_counties_from_fp,
)
from elm.ords.utilities.location import County
from elm.ords.utilities.queued_logging import (
    LocationFileLog,
    LogListener,
    NoLocationFilter,
)

logger = logging.getLogger(__name__)


OUT_COLS = [
    "county",
    "state",
    "FIPS",
    "feature",
    "fixed_value",
    "mult_value",
    "mult_type",
    "adder",
    "min_dist",
    "max_dist",
    "value",
    "units",
    "ord_year",
    "last_updated",
    "section",
    "source",
    "comment",
]

CHECK_COLS = [
    "fixed_value",
    "mult_value",
    "adder",
    "min_dist",
    "max_dist",
    "value",
]


async def process_counties_with_openai(
    out_dir,
    county_fp=None,
    model="gpt-4",
    azure_api_key=None,
    azure_version=None,
    azure_endpoint=None,
    llm_call_kwargs=None,
    llm_service_rate_limit=4000,
    text_splitter_chunk_size=3000,
    text_splitter_chunk_overlap=300,
    num_urls_to_check_per_county=5,
    max_num_concurrent_browsers=10,
    file_loader_kwargs=None,
    pytesseract_exe_fp=None,
    td_kwargs=None,
    tpe_kwargs=None,
    ppe_kwargs=None,
    log_dir=None,
    clean_dir=None,
    county_ords_dir=None,
    county_dbs_dir=None,
    log_level="INFO",
):
    """Download and extract ordinances for a list of counties.

    Parameters
    ----------
    out_dir : path-like
        Path to output directory. This directory will be created if it
        does not exist. This directory will contain the structured
        ordinance output CSV as well as all of the scraped ordinance
        documents (PDFs and HTML text files). Usage information and
        default options for log/clean directories will also be stored
        here.
    county_fp : path-like, optional
        Path to CSV file containing a list of counties to extract
        ordinance information for. This CSV should have "County" and
        "State" columns that contains the county and state names.
        By default, ``None``, which runs the extraction for all known
        counties (this is untested and not currently recommended).
    model : str, optional
        Name of LLM model to perform scraping. By default, ``"gpt-4"``.
    azure_api_key : str, optional
        Azure OpenAI API key. By default, ``None``, which pulls the key
        from the environment variable ``AZURE_OPENAI_API_KEY`` instead.
    azure_version : str, optional
        Azure OpenAI API version. By default, ``None``, which pulls the
        version from the environment variable ``AZURE_OPENAI_VERSION``
        instead.
    azure_endpoint : str, optional
        Azure OpenAI API endpoint. By default, ``None``, which pulls the
        endpoint from the environment variable ``AZURE_OPENAI_ENDPOINT``
        instead.
    llm_call_kwargs : dict, optional
        Keyword-value pairs used to initialize an
        `elm.ords.llm.LLMCaller` instance. By default, ``None``.
    llm_service_rate_limit : int, optional
        Token rate limit of LLm service being used (OpenAI).
        By default, ``4000``.
    text_splitter_chunk_size : int, optional
        Chunk size input to
        `langchain.text_splitter.RecursiveCharacterTextSplitter`.
        By default, ``3000``.
    text_splitter_chunk_overlap : int, optional
        Chunk overlap input to
        `langchain.text_splitter.RecursiveCharacterTextSplitter`.
        By default, ``300``.
    num_urls_to_check_per_county : int, optional
        Number of unique Google search result URL's to check for
        ordinance document. By default, ``5``.
    max_num_concurrent_browsers : int, optional
        Number of unique concurrent browser instances to open when
        performing Google search. Setting this number too high on a
        machine with limited processing can lead to increased timeouts
        and therefore decreased quality of Google search results.
        By default, ``10``.
    pytesseract_exe_fp : path-like, optional
        Path to pytesseract executable. If this option is specified, OCR
        parsing for PDf files will be enabled via pytesseract.
        By default, ``None``.
    td_kwargs : dict, optional
        Keyword-value argument pairs to pass to
        :class:`tempfile.TemporaryDirectory`. The temporary directory is
        used to store files downloaded from the web that are still being
        parsed for ordinance information. By default, ``None``.
    tpe_kwargs : dict, optional
        Keyword-value argument pairs to pass to
        :class:`concurrent.futures.ThreadPoolExecutor`. The thread pool
        executor is used to run I/O intensive tasks like writing to a
        log file. By default, ``None``.
    ppe_kwargs : dict, optional
        Keyword-value argument pairs to pass to
        :class:`concurrent.futures.ProcessPoolExecutor`. The process
        pool executor is used to run CPU intensive tasks like loading
        a PDF file. By default, ``None``.
    log_dir : path-like, optional
        Path to directory for log files. This directory will be created
        if it does not exist. By default, ``None``, which
        creates a ``logs`` folder in the output directory for the
        county-specific log files.
    clean_dir : path-like, optional
        Path to directory for cleaned ordinance text output. This
        directory will be created if it does not exist. By default,
        ``None``, which creates a ``clean`` folder in the output
        directory for the cleaned ordinance text files.
    county_ords_dir : path-like, optional
        Path to directory for individual county ordinance file outputs.
        This directory will be created if it does not exist.
        By default, ``None``, which creates a ``county_ord_files``
        folder in the output directory.
    county_dbs_dir : path-like, optional
        Path to directory for individual county ordinance database
        outputs. This directory will be created if it does not exist.
        By default, ``None``, which creates a ``county_dbs`` folder in
        the output directory.
    log_level : str, optional
        Log level to set for county retrieval and parsing loggers.
        By default, ``"INFO"``.

    Returns
    -------
    pd.DataFrame
        DataFrame of parsed ordinance information. This file will also
        be stored in the output directory under "wind_db.csv".
    """
    start_time = time.time()
    log_listener = LogListener(["elm"], level=log_level)
    dirs = _setup_folders(
        out_dir,
        log_dir=log_dir,
        clean_dir=clean_dir,
        cod=county_ords_dir,
        cdd=county_dbs_dir,
    )
    out_dir, log_dir, clean_dir, county_ords_dir, county_dbs_dir = dirs
    async with log_listener as ll:
        _setup_main_logging(log_dir, log_level, ll)
        db = await _process_with_logs(
            out_dir,
            log_dir,
            clean_dir,
            county_ords_dir,
            county_dbs_dir,
            ll,
            county_fp=county_fp,
            model=model,
            azure_api_key=azure_api_key,
            azure_version=azure_version,
            azure_endpoint=azure_endpoint,
            llm_call_kwargs=llm_call_kwargs,
            llm_service_rate_limit=llm_service_rate_limit,
            text_splitter_chunk_size=text_splitter_chunk_size,
            text_splitter_chunk_overlap=text_splitter_chunk_overlap,
            num_urls_to_check_per_county=num_urls_to_check_per_county,
            max_num_concurrent_browsers=max_num_concurrent_browsers,
            file_loader_kwargs=file_loader_kwargs,
            pytesseract_exe_fp=pytesseract_exe_fp,
            td_kwargs=td_kwargs,
            tpe_kwargs=tpe_kwargs,
            ppe_kwargs=ppe_kwargs,
            log_level=log_level,
        )
    _record_total_time(out_dir / "usage.json", time.time() - start_time)
    return db


async def _process_with_logs(
    out_dir,
    log_dir,
    clean_dir,
    county_ords_dir,
    county_dbs_dir,
    log_listener,
    county_fp=None,
    model="gpt-4",
    azure_api_key=None,
    azure_version=None,
    azure_endpoint=None,
    llm_call_kwargs=None,
    llm_service_rate_limit=4000,
    text_splitter_chunk_size=3000,
    text_splitter_chunk_overlap=300,
    num_urls_to_check_per_county=5,
    max_num_concurrent_browsers=10,
    file_loader_kwargs=None,
    pytesseract_exe_fp=None,
    td_kwargs=None,
    tpe_kwargs=None,
    ppe_kwargs=None,
    log_level="INFO",
):
    """Process counties with logging enabled."""
    counties = _load_counties_to_process(county_fp)
    azure_api_key, azure_version, azure_endpoint = validate_azure_api_params(
        azure_api_key, azure_version, azure_endpoint
    )

    tpe_kwargs = _configure_thread_pool_kwargs(tpe_kwargs)
    file_loader_kwargs = _configure_file_loader_kwargs(file_loader_kwargs)
    if pytesseract_exe_fp is not None:
        _setup_pytesseract(pytesseract_exe_fp)
        file_loader_kwargs.update({"pdf_ocr_read_coroutine": read_pdf_doc_ocr})

    text_splitter = RecursiveCharacterTextSplitter(
        RTS_SEPARATORS,
        chunk_size=text_splitter_chunk_size,
        chunk_overlap=text_splitter_chunk_overlap,
        length_function=partial(ApiBase.count_tokens, model=model),
    )
    client = openai.AsyncAzureOpenAI(
        api_key=azure_api_key,
        api_version=azure_version,
        azure_endpoint=azure_endpoint,
    )

    services = [
        OpenAIService(client, rate_limit=llm_service_rate_limit),
        TempFileCache(td_kwargs=td_kwargs, tpe_kwargs=tpe_kwargs),
        FileMover(county_ords_dir, tpe_kwargs=tpe_kwargs),
        CleanedFileWriter(clean_dir, tpe_kwargs=tpe_kwargs),
        OrdDBFileWriter(county_dbs_dir, tpe_kwargs=tpe_kwargs),
        UsageUpdater(out_dir / "usage.json", tpe_kwargs=tpe_kwargs),
        PDFLoader(**(ppe_kwargs or {})),
    ]

    browser_semaphore = (
        asyncio.Semaphore(max_num_concurrent_browsers)
        if max_num_concurrent_browsers
        else None
    )

    async with RunningAsyncServices(services):
        tasks = []
        trackers = []
        for __, row in counties.iterrows():
            county, state, fips = row[["County", "State", "FIPS"]]
            location = County(county.strip(), state=state.strip(), fips=fips)
            usage_tracker = UsageTracker(
                location.full_name, usage_from_response
            )
            trackers.append(usage_tracker)
            task = asyncio.create_task(
                process_county_with_logging(
                    log_listener,
                    log_dir,
                    location,
                    text_splitter,
                    num_urls=num_urls_to_check_per_county,
                    file_loader_kwargs=file_loader_kwargs,
                    browser_semaphore=browser_semaphore,
                    level=log_level,
                    llm_service=OpenAIService,
                    usage_tracker=usage_tracker,
                    model=model,
                    **(llm_call_kwargs or {}),
                ),
                name=location.full_name,
            )
            tasks.append(task)
        docs = await asyncio.gather(*tasks)

    db = _docs_to_db(docs)
    db.to_csv(out_dir / "wind_db.csv", index=False)
    return db


def _setup_main_logging(log_dir, level, listener):
    """Setup main logger for catching exceptions during execution."""
    handler = logging.FileHandler(log_dir / "main.log", encoding="utf-8")
    handler.setLevel(level)
    handler.addFilter(NoLocationFilter())
    listener.addHandler(handler)


def _setup_folders(
    out_dir,
    log_dir=None,
    clean_dir=None,
    cod=None,
    cdd=None,
):
    """Setup output directory folders."""
    out_dir = Path(out_dir)
    out_folders = [
        out_dir,
        Path(log_dir) if log_dir else out_dir / "logs",
        Path(clean_dir) if clean_dir else out_dir / "clean",
        Path(cod) if cod else out_dir / "county_ord_files",
        Path(cdd) if cdd else out_dir / "county_dbs",
    ]
    for folder in out_folders:
        folder.mkdir(exist_ok=True, parents=True)
    return out_folders


def _load_counties_to_process(county_fp):
    """Load the counties to retrieve documents for."""
    if county_fp is None:
        logger.info("No `county_fp` input! Loading all counties")
        return load_all_county_info()
    return load_counties_from_fp(county_fp)


def _configure_thread_pool_kwargs(tpe_kwargs):
    """Set thread pool workers to 5 if user didn't specify."""
    tpe_kwargs = tpe_kwargs or {}
    tpe_kwargs.setdefault("max_workers", 5)
    return tpe_kwargs


def _configure_file_loader_kwargs(file_loader_kwargs):
    """Add PDF reading coroutine to kwargs."""
    file_loader_kwargs = file_loader_kwargs or {}
    file_loader_kwargs.update({"pdf_read_coroutine": read_pdf_doc})
    return file_loader_kwargs


async def process_county_with_logging(
    listener,
    log_dir,
    county,
    text_splitter,
    num_urls=5,
    file_loader_kwargs=None,
    browser_semaphore=None,
    level="INFO",
    **kwargs,
):
    """Retrieve ordinance document for a single county with async logs.

    Parameters
    ----------
    listener : elm.ords.utilities.queued_logging.LogListener
        Active ``LogListener`` instance that can be passed to
        :class:`elm.ords.utilities.queued_logging.LocationFileLog`.
    log_dir : path-like
        Path to output directory to contain log file.
    county : elm.ords.utilities.location.Location
        County to retrieve ordinance document for.
    text_splitter : obj, optional
        Instance of an object that implements a `split_text` method.
        The method should take text as input (str) and return a list
        of text chunks. Langchain's text splitters should work for this
        input.
    num_urls : int, optional
        Number of unique Google search result URL's to check for
        ordinance document. By default, ``5``.
    file_loader_kwargs : dict, optional
        Dictionary of keyword-argument pairs to initialize
        :class:`elm.web.file_loader.AsyncFileLoader` with. The
        "pw_launch_kwargs" key in these will also be used to initialize
        the :class:`elm.web.google_search.PlaywrightGoogleLinkSearch`
        used for the google URL search. By default, ``None``.
    browser_semaphore : asyncio.Semaphore, optional
        Semaphore instance that can be used to limit the number of
        playwright browsers open concurrently. If ``None``, no limits
        are applied. By default, ``None``.
    level : str, optional
        Log level to set for retrieval logger. By default, ``"INFO"``.
    **kwargs
        Keyword-value pairs used to initialize an
        `elm.ords.llm.LLMCaller` instance.

    Returns
    -------
    elm.web.document.BaseDocument | None
        Document instance for the ordinance document, or ``None`` if no
        document was found. Extracted ordinance information is stored in
        the document's ``metadata`` attribute.
    """
    with LocationFileLog(
        listener, log_dir, location=county.full_name, level=level
    ):
        task = asyncio.create_task(
            process_county(
                county,
                text_splitter,
                num_urls=num_urls,
                file_loader_kwargs=file_loader_kwargs,
                browser_semaphore=browser_semaphore,
                **kwargs,
            ),
            name=county.full_name,
        )
        try:
            doc, *__ = await asyncio.gather(task)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            logger.error(
                "Encountered error while processing %s:", county.full_name
            )
            logger.exception(e)
            doc = None

        return doc


async def process_county(
    county,
    text_splitter,
    num_urls=5,
    file_loader_kwargs=None,
    browser_semaphore=None,
    **kwargs,
):
    """Download and parse ordinance document for a single county.

    Parameters
    ----------
    county : elm.ords.utilities.location.Location
        County to retrieve ordinance document for.
    text_splitter : obj, optional
        Instance of an object that implements a `split_text` method.
        The method should take text as input (str) and return a list
        of text chunks. Langchain's text splitters should work for this
        input.
    num_urls : int, optional
        Number of unique Google search result URL's to check for
        ordinance document. By default, ``5``.
    file_loader_kwargs : dict, optional
        Dictionary of keyword-argument pairs to initialize
        :class:`elm.web.file_loader.AsyncFileLoader` with. The
        "pw_launch_kwargs" key in these will also be used to initialize
        the :class:`elm.web.google_search.PlaywrightGoogleLinkSearch`
        used for the google URL search. By default, ``None``.
    browser_semaphore : asyncio.Semaphore, optional
        Semaphore instance that can be used to limit the number of
        playwright browsers open concurrently. If ``None``, no limits
        are applied. By default, ``None``.
    **kwargs
        Keyword-value pairs used to initialize an
        `elm.ords.llm.LLMCaller` instance.

    Returns
    -------
    elm.web.document.BaseDocument | None
        Document instance for the ordinance document, or ``None`` if no
        document was found. Extracted ordinance information is stored in
        the document's ``metadata`` attribute.
    """
    start_time = time.time()
    doc = await download_county_ordinance(
        county,
        text_splitter,
        num_urls=num_urls,
        file_loader_kwargs=file_loader_kwargs,
        browser_semaphore=browser_semaphore,
        **kwargs,
    )
    if doc is None:
        await _record_time_and_usage(start_time, **kwargs)
        return None

    doc.attrs["location"] = county
    doc.attrs["location_name"] = county.full_name
    await _record_usage(**kwargs)

    doc = await extract_ordinance_text_with_ngram_validation(
        doc, text_splitter, **kwargs
    )
    await _record_usage(**kwargs)

    doc = await _write_cleaned_text(doc)
    doc = await extract_ordinance_values(doc, **kwargs)

    ord_count = _num_ords_in_doc(doc)
    if ord_count > 0:
        doc = await _move_file_to_out_dir(doc)
        doc = await _write_ord_db(doc)
        logger.info(
            "%d ordinance value(s) found for %s. Outputs are here: '%s'",
            ord_count,
            county.full_name,
            doc.attrs["ord_db_fp"],
        )
    else:
        logger.info("No ordinances found for %s.", county.full_name)

    await _record_time_and_usage(start_time, **kwargs)
    return doc


async def _record_usage(**kwargs):
    """Dump usage to file if tracker found in kwargs."""
    usage_tracker = kwargs.get("usage_tracker")
    if usage_tracker:
        await UsageUpdater.call(usage_tracker)


async def _record_time_and_usage(start_time, **kwargs):
    """Add elapsed time before updating usage to file."""
    seconds_elapsed = time.time() - start_time
    usage_tracker = kwargs.get("usage_tracker")
    if usage_tracker:
        usage_tracker["total_time_seconds"] = seconds_elapsed
        usage_tracker["total_time"] = str(timedelta(seconds=seconds_elapsed))
        await UsageUpdater.call(usage_tracker)


async def _move_file_to_out_dir(doc):
    """Move PDF or HTML text file to output directory."""
    out_fp = await FileMover.call(doc)
    doc.attrs["out_fp"] = out_fp
    return doc


async def _write_cleaned_text(doc):
    """Write cleaned text to `clean_dir`."""
    out_fp = await CleanedFileWriter.call(doc)
    doc.attrs["cleaned_fp"] = out_fp
    return doc


async def _write_ord_db(doc):
    """Write cleaned text to `county_dbs_dir`."""
    out_fp = await OrdDBFileWriter.call(doc)
    doc.attrs["ord_db_fp"] = out_fp
    return doc


def _setup_pytesseract(exe_fp):
    """Set the pytesseract command."""
    import pytesseract

    logger.debug("Setting `tesseract_cmd` to %s", exe_fp)
    pytesseract.pytesseract.tesseract_cmd = exe_fp


def _record_total_time(fp, seconds_elapsed):
    """Dump usage to an existing file."""
    if not Path(fp).exists():
        usage_info = {}
    else:
        with open(fp, "r") as fh:
            usage_info = json.load(fh)

    total_time_str = str(timedelta(seconds=seconds_elapsed))
    usage_info["total_time_seconds"] = seconds_elapsed
    usage_info["total_time"] = total_time_str

    with open(fp, "w") as fh:
        json.dump(usage_info, fh, indent=4)

    logger.info("Total processing time: %s", total_time_str)


def _num_ords_in_doc(doc):
    """Check if doc contains any scraped ordinance values."""
    if doc is None:
        return 0

    if "ordinance_values" not in doc.attrs:
        return 0

    ord_vals = doc.attrs["ordinance_values"]
    if ord_vals.empty:
        return 0

    check_cols = [col for col in CHECK_COLS if col in ord_vals]
    if not check_cols:
        return 0

    return (~ord_vals[check_cols].isna()).values.sum(axis=1).sum()


def _docs_to_db(docs):
    """Convert list of docs to output database."""
    db = []
    for doc in docs:
        if doc is None or isinstance(doc, Exception):
            continue

        if _num_ords_in_doc(doc) == 0:
            continue

        results = _db_results(doc)
        results = _formatted_db(results)
        db.append(results)

    if not db:
        return pd.DataFrame(columns=OUT_COLS)

    db = pd.concat(db)
    db = _empirical_adjustments(db)
    return _formatted_db(db)


def _db_results(doc):
    """Extract results from doc metadata to DataFrame."""
    results = doc.attrs.get("ordinance_values")
    if results is None:
        return None

    results["source"] = doc.attrs.get("source")
    year = doc.attrs.get("date", (None, None, None))[0]
    results["ord_year"] = year if year is not None and year > 0 else None
    results["last_updated"] = datetime.now().strftime("%m/%d/%Y")

    location = doc.attrs["location"]
    results["FIPS"] = location.fips
    results["county"] = location.name
    results["state"] = location.state
    return results


def _empirical_adjustments(db):
    """Post-processing adjustments based on empirical observations.

    Current adjustments include:

        - Limit adder to max of 250 ft.
            - Chat GPT likes to report large values here, but in
            practice all values manually observed in ordinance documents
            are below 250 ft.

    """
    if "adder" in db.columns:
        db.loc[db["adder"] > 250, "adder"] = None
    return db


def _formatted_db(db):
    """Format DataFrame for output."""
    out_cols = [col for col in OUT_COLS if col in db.columns]
    return db[out_cols]



# -*- coding: utf-8 -*-
"""ELM Web file loader class."""
import asyncio
import logging

import aiohttp
from fake_useragent import UserAgent

from elm.utilities.parse import read_pdf
from elm.web.document import PDFDocument, HTMLDocument
from elm.web.html_pw import load_html_with_pw
from elm.web.utilities import DEFAULT_HEADERS
from elm.utilities.retry import async_retry_with_exponential_backoff
from elm.exceptions import ELMRuntimeError


logger = logging.getLogger(__name__)


async def _read_pdf_doc(pdf_bytes, **kwargs):
    """Default read PDF function (runs in main thread)"""
    verbose = kwargs.pop("verbose", True)
    pages = read_pdf(pdf_bytes, verbose=verbose)
    return PDFDocument(pages, **kwargs)


async def _read_html_doc(text, **kwargs):
    """Default read HTML function (runs in main thread)"""
    return HTMLDocument([text], **kwargs)


class AsyncFileLoader:
    """Async web file (PDF or HTML) loader

    Purpose:
        Save content from links as files.
    Responsibilities:
        1. Retrieve data from a URL.
        2. Determine wether information should be stored as a PDF or
           HTML document.
    Key Relationships:
        Returns either :class:`~elm.web.document.PDFDocument` or
        :class:`~elm.web.document.HTMLDocument`. Uses `aiohttp` to
        access the web.

    .. end desc
    """

    PAGE_LOAD_TIMEOUT = 60_000
    """Default page load timeout value in milliseconds"""

    def __init__(
        self,
        header_template=None,
        verify_ssl=True,
        aget_kwargs=None,
        pw_launch_kwargs=None,
        pdf_read_kwargs=None,
        html_read_kwargs=None,
        pdf_read_coroutine=None,
        html_read_coroutine=None,
        pdf_ocr_read_coroutine=None,
        file_cache_coroutine=None,
        browser_semaphore=None,
        use_scrapling_stealth=False,
        **__,  # consume any extra kwargs
    ):
        """

        Parameters
        ----------
        header_template : dict, optional
            Optional GET header template. If not specified, uses
            :obj:`~elm.web.utilities.DEFAULT_HEADERS`.
            By default, ``None``.
        verify_ssl : bool, optional
            Option to use aiohttp's default SSL check. If ``False``,
            SSL certificate validation is skipped. By default, ``True``.
        aget_kwargs : dict, optional
            Other kwargs to pass to :meth:`aiohttp.ClientSession.get`.
            By default, ``None``.
        pw_launch_kwargs : dict, optional
            Keyword-value argument pairs to pass to
            :meth:`async_playwright.chromium.launch` (only used when
            reading HTML). By default, ``None``.
        pdf_read_kwargs : dict, optional
            Keyword-value argument pairs to pass to the
            `pdf_read_coroutine`. By default, ``None``.
        html_read_kwargs : dict, optional
            Keyword-value argument pairs to pass to the
            `html_read_coroutine`. By default, ``None``.
        pdf_read_coroutine : callable, optional
            PDF file read coroutine. Must by an async function. Should
            accept PDF bytes as the first argument and kwargs as the
            rest. Must return a :obj:`elm.web.document.PDFDocument`.
            If ``None``, a default function that runs in the main thread
            is used. By default, ``None``.
        html_read_coroutine : callable, optional
            HTML file read coroutine. Must by an async function. Should
            accept HTML text as the first argument and kwargs as the
            rest. Must return a :obj:`elm.web.document.HTMLDocument`.
            If ``None``, a default function that runs in the main thread
            is used. By default, ``None``.
        pdf_ocr_read_coroutine : callable, optional
            PDF OCR file read coroutine. Must by an async function.
            Should accept PDF bytes as the first argument and kwargs as
            the rest. Must return a :obj:`elm.web.document.PDFDocument`.
            If ``None``, PDF OCR parsing is not attempted, and any
            scanned PDF URL's will return a blank document.
            By default, ``None``.
        file_cache_coroutine : callable, optional
            File caching coroutine. Can be used to cache files
            downloaded by this class. Must accept an
            :obj:`~elm.web.document.Document` instance as the first
            argument and the file content to be written as the second
            argument. If this method is not provided, no document
            caching is performed. By default, ``None``.
        browser_semaphore : asyncio.Semaphore, optional
            Semaphore instance that can be used to limit the number of
            playwright browsers open concurrently. If ``None``, no
            limits are applied. By default, ``None``.
        use_scrapling_stealth : bool, default=False
            Option to use scrapling stealth scripts instead of
            tf-playwright-stealth. By default, ``False``.
        """
        self.pw_launch_kwargs = pw_launch_kwargs or {}
        self.pdf_read_kwargs = pdf_read_kwargs or {}
        self.html_read_kwargs = html_read_kwargs or {}
        self.get_kwargs = {
            "headers": self._header_from_template(header_template),
            "ssl": None if verify_ssl else False,
            **(aget_kwargs or {}),
        }
        self.pdf_read_coroutine = pdf_read_coroutine or _read_pdf_doc
        self.html_read_coroutine = html_read_coroutine or _read_html_doc
        self.pdf_ocr_read_coroutine = pdf_ocr_read_coroutine
        self.file_cache_coroutine = file_cache_coroutine
        self.browser_semaphore = browser_semaphore
        self.uss = use_scrapling_stealth

    def _header_from_template(self, header_template):
        """Compile header from user or default template"""
        headers = header_template or DEFAULT_HEADERS
        headers = dict(headers)
        if not headers.get("User-Agent"):
            headers["User-Agent"] = UserAgent().random
        return headers

    async def fetch_all(self, *urls):
        """Fetch documents for all requested URL's.

        Parameters
        ----------
        *urls
            Iterable of URL's (as strings) to fetch.

        Returns
        -------
        list
            List of documents, one per requested URL.
        """
        outer_task_name = asyncio.current_task().get_name()
        fetches = [
            asyncio.create_task(self.fetch(url), name=outer_task_name)
            for url in urls
        ]
        return await asyncio.gather(*fetches)

    async def fetch(self, url):
        """Fetch a document for the given URL.

        Parameters
        ----------
        url : str
            URL for the document to pull down.

        Returns
        -------
        :class:`elm.web.document.Document`
            Document instance containing text, if the fetch was
            successful.
        """
        try:
            doc, raw_content = await self._fetch_doc_with_url_in_metadata(url)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            msg = ("Encountered error of type %r while fetching document from "
                   "%s:")
            err_type = type(e)
            logger.exception(msg, err_type, url)
            return HTMLDocument(pages=[])

        doc = await self._cache_doc(doc, raw_content)
        return doc

    async def _fetch_doc_with_url_in_metadata(self, url):
        """Fetch doc contents and add URL to metadata"""
        doc, raw_content = await self._fetch_doc(url)
        doc.attrs["source"] = url
        return doc, raw_content

    async def _fetch_doc(self, url):
        """Fetch a doc by trying pdf read, then HTML read, then PDF OCR"""

        async with aiohttp.ClientSession() as session:
            try:
                logger.debug("Fetching content from %r", url)
                url_bytes = await self._fetch_content_with_retry(url, session)
            except ELMRuntimeError:
                logger.exception("Could not fetch content from %r", url)
                return PDFDocument(pages=[]), None

        logger.debug("Got content from %r", url)
        doc = await self.pdf_read_coroutine(url_bytes, **self.pdf_read_kwargs)
        if not doc.empty:
            return doc, url_bytes

        logger.debug("PDF read failed; fetching HTML content from %r", url)
        text = await load_html_with_pw(url, self.browser_semaphore,
                                       timeout=self.PAGE_LOAD_TIMEOUT,
                                       use_scrapling_stealth=self.uss,
                                       **self.pw_launch_kwargs)
        doc = await self.html_read_coroutine(text, **self.html_read_kwargs)
        if not doc.empty:
            return doc, doc.text

        if self.pdf_ocr_read_coroutine:
            logger.debug("HTML read failed; fetching OCR content from %r", url)
            doc = await self.pdf_ocr_read_coroutine(
                url_bytes, **self.pdf_read_kwargs
            )

        return doc, url_bytes

    @async_retry_with_exponential_backoff(
        base_delay=2,
        exponential_base=1.5,
        jitter=False,
        max_retries=3,
        errors=(
            aiohttp.ClientConnectionError,
            aiohttp.client_exceptions.ClientError,
        ),
    )
    async def _fetch_content_with_retry(self, url, session):
        """Fetch content from URL with several retry attempts"""
        async with session.get(url, **self.get_kwargs) as response:
            return await response.read()

    async def _cache_doc(self, doc, raw_content):
        """Cache doc if user provided a coroutine"""
        if doc.empty or not raw_content:
            return doc

        if not self.file_cache_coroutine:
            return doc

        cache_fn = await self.file_cache_coroutine(doc, raw_content)
        if cache_fn is not None:
            doc.attrs["cache_fn"] = cache_fn
        return doc



# -*- coding: utf-8 -*-
"""ELM Ordinance document content Validation logic

These are primarily used to validate that a legal document applies to a
particular technology (e.g. Large Wind Energy Conversion Systems).
"""
import logging


logger = logging.getLogger(__name__)
NOT_WIND_WORDS = [
    "windy",
    "winds",
    "window",
    "windiest",
    "windbreak",
    "windshield",
    "wind blow",
    "wind erosion",
    "rewind",
    "mini wecs",
    "swecs",
    "private wecs",
    "pwecs",
    "wind direction",
    "wind movement",
    "wind attribute",
    "wind runway",
    "wind load",
    "wind orient",
    "wind damage",
]
GOOD_WIND_KEYWORDS = ["wind", "setback"]
GOOD_WIND_ACRONYMS = ["wecs", "wes", "lwet", "uwet", "wef"]
_GOOD_ACRONYM_CONTEXTS = [
    " {acronym} ",
    " {acronym}\n",
    " {acronym}.",
    "\n{acronym} ",
    "\n{acronym}.",
    "\n{acronym}\n",
    "({acronym} ",
    " {acronym})",
]
GOOD_WIND_PHRASES = ["wind energy conversion", "wind turbine", "wind tower"]


class ValidationWithMemory:
    """Validate a set of text chunks by sometimes looking at previous chunks"""

    def __init__(self, structured_llm_caller, text_chunks, num_to_recall=2):
        """

        Parameters
        ----------
        structured_llm_caller : elm.ords.llm.StructuredLLMCaller
            StructuredLLMCaller instance. Used for structured validation
            queries.
        text_chunks : list of str
            List of strings, each of which represent a chunk of text.
            The order of the strings should be the order of the text
            chunks. This validator may refer to previous text chunks to
            answer validation questions.
        num_to_recall : int, optional
            Number of chunks to check for each validation call. This
            includes the original chunk! For example, if
            `num_to_recall=2`, the validator will first check the chunk
            at the requested index, and then the previous chunk as well.
            By default, ``2``.
        """
        self.slc = structured_llm_caller
        self.text_chunks = text_chunks
        self.num_to_recall = num_to_recall
        self.memory = [{} for _ in text_chunks]

    # fmt: off
    def _inverted_mem(self, starting_ind):
        """Inverted memory."""
        inverted_mem = self.memory[:starting_ind + 1:][::-1]
        yield from inverted_mem[:self.num_to_recall]

    # fmt: off
    def _inverted_text(self, starting_ind):
        """Inverted text chunks"""
        inverted_text = self.text_chunks[:starting_ind + 1:][::-1]
        yield from inverted_text[:self.num_to_recall]

    async def parse_from_ind(self, ind, prompt, key):
        """Validate a chunk of text.

        Validation occurs by querying the LLM using the input prompt and
        parsing the `key` from the response JSON. The prompt should
        request that the key be a boolean output. If the key retrieved
        from the LLM response is False, a number of previous text chunks
        are checked as well, using the same prompt. This can be helpful
        in cases where the answer to the validation prompt (e.g. does
        this text pertain to a large WECS?) is only found in a previous
        text chunk.

        Parameters
        ----------
        ind : int
            Positive integer corresponding to the chunk index.
            Must be less than `len(text_chunks)`.
        prompt : str
            Input LLM system prompt that describes the validation
            question. This should request a JSON output from the LLM.
            It should also take `key` as a formatting input.
        key : str
            A key expected in the JSON output of the LLM containing the
            response for the validation question. This string will also
            be used to format the system prompt before it is passed to
            the LLM.

        Returns
        -------
        bool
            ``True`` if the LLM returned ``True`` for this text chunk or
            `num_to_recall-1` text chunks before it.
            ``False`` otherwise.
        """
        logger.debug("Checking %r for ind %d", key, ind)
        mem_text = zip(self._inverted_mem(ind), self._inverted_text(ind))
        for step, (mem, text) in enumerate(mem_text):
            logger.debug("Mem at ind %d is %s", step, mem)
            check = mem.get(key)
            if check is None:
                content = await self.slc.call(
                    sys_msg=prompt.format(key=key),
                    content=text,
                    usage_sub_label="document_content_validation",
                )
                check = mem[key] = content.get(key, False)
            if check:
                return check
        return False


def possibly_mentions_wind(text, match_count_threshold=1):
    """Perform a heuristic check for mention of wind energy in text.

    This check first strips the text of any wind "look-alike" words
    (e.g. "window", "windshield", etc). Then, it checks for particular
    keywords, acronyms, and phrases that pertain to wind in the text.
    If enough keywords are mentions (as dictated by
    `match_count_threshold`), this check returns ``True``.

    Parameters
    ----------
    text : str
        Input text that may or may not mention win in relation to wind
        energy.
    match_count_threshold : int, optional
        Number of keywords that must match for the text to pass this
        heuristic check. Count must be strictly greater than this value.
        By default, ``1``.

    Returns
    -------
    bool
        ``True`` if the number of keywords/acronyms/phrases detected
        exceeds the `match_count_threshold`.
    """
    heuristics_text = _convert_to_heuristics_text(text)
    total_keyword_matches = _count_single_keyword_matches(heuristics_text)
    total_keyword_matches += _count_acronym_matches(heuristics_text)
    total_keyword_matches += _count_phrase_matches(heuristics_text)
    return total_keyword_matches > match_count_threshold


def _convert_to_heuristics_text(text):
    """Convert text for heuristic wind content parsing"""
    heuristics_text = text.casefold()
    for word in NOT_WIND_WORDS:
        heuristics_text = heuristics_text.replace(word, "")
    return heuristics_text


def _count_single_keyword_matches(heuristics_text):
    """Count number of good wind energy keywords that appear in text."""
    return sum(keyword in heuristics_text for keyword in GOOD_WIND_KEYWORDS)


def _count_acronym_matches(heuristics_text):
    """Count number of good wind energy acronyms that appear in text."""
    acronym_matches = 0
    for context in _GOOD_ACRONYM_CONTEXTS:
        acronym_keywords = {
            context.format(acronym=acronym) for acronym in GOOD_WIND_ACRONYMS
        }
        acronym_matches = sum(
            keyword in heuristics_text for keyword in acronym_keywords
        )
        if acronym_matches > 0:
            break
    return acronym_matches


def _count_phrase_matches(heuristics_text):
    """Count number of good wind energy phrases that appear in text."""
    return sum(
        all(keyword in heuristics_text for keyword in phrase.split(" "))
        for phrase in GOOD_WIND_PHRASES
    )



# -*- coding: utf-8 -*-
"""ELM Web Scraping - functions to run a search"""
import pprint
import asyncio
import random
import logging
from collections import namedtuple
from itertools import zip_longest, chain
from contextlib import AsyncExitStack

from elm.web.file_loader import AsyncFileLoader
from elm.web.search.bing import PlaywrightBingLinkSearch
from elm.web.search.duckduckgo import (APIDuckDuckGoSearch,
                                       PlaywrightDuckDuckGoLinkSearch)
from elm.web.search.dux import DuxDistributedGlobalSearch
from elm.web.search.google import (APIGoogleCSESearch, APISerperSearch,
                                   SerpAPIGoogleSearch,
                                   CamoufoxGoogleLinkSearch,
                                   PlaywrightGoogleCSELinkSearch,
                                   PlaywrightGoogleLinkSearch)
from elm.web.search.tavily import APITavilySearch
from elm.web.search.yahoo import PlaywrightYahooLinkSearch
from elm.exceptions import ELMKeyError, ELMInputError


logger = logging.getLogger(__name__)


_RESULTS_PER_QUERY = 10
_SE_OPT = namedtuple('_SE_OPT', ['se_class', 'uses_browser', 'kwg_key_name'])
SEARCH_ENGINE_OPTIONS = {
    "APIDuckDuckGoSearch": _SE_OPT(APIDuckDuckGoSearch, False,
                                   "ddg_api_kwargs"),
    "APIGoogleCSESearch": _SE_OPT(APIGoogleCSESearch, False,
                                  "google_cse_api_kwargs"),
    "APISerperSearch": _SE_OPT(APISerperSearch, False,
                               "google_serper_api_kwargs"),
    "SerpAPIGoogleSearch": _SE_OPT(SerpAPIGoogleSearch, False,
                                   "google_serpapi_kwargs"),
    "APITavilySearch": _SE_OPT(APITavilySearch, False, "tavily_api_kwargs"),
    "CamoufoxGoogleLinkSearch": _SE_OPT(CamoufoxGoogleLinkSearch, True,
                                        "cf_google_se_kwargs"),
    "DuxDistributedGlobalSearch": _SE_OPT(DuxDistributedGlobalSearch, False,
                                          "ddgs_kwargs"),
    "PlaywrightBingLinkSearch": _SE_OPT(PlaywrightBingLinkSearch, True,
                                        "pw_bing_se_kwargs"),
    "PlaywrightDuckDuckGoLinkSearch": _SE_OPT(PlaywrightDuckDuckGoLinkSearch,
                                              True, "pw_ddg_se_kwargs"),
    "PlaywrightGoogleCSELinkSearch": _SE_OPT(PlaywrightGoogleCSELinkSearch,
                                             True, "pw_google_cse_kwargs"),
    "PlaywrightGoogleLinkSearch": _SE_OPT(PlaywrightGoogleLinkSearch, True,
                                          "pw_google_se_kwargs"),
    "PlaywrightYahooLinkSearch": _SE_OPT(PlaywrightYahooLinkSearch, True,
                                         "pw_yahoo_se_kwargs")
}
"""Supported search engines"""
_DEFAULT_SE = ("PlaywrightGoogleLinkSearch", "PlaywrightDuckDuckGoLinkSearch",
               "DuxDistributedGlobalSearch")


async def web_search_links_as_docs(queries, search_engines=_DEFAULT_SE,
                                   num_urls=None, ignore_url_parts=None,
                                   search_semaphore=None,
                                   browser_semaphore=None, task_name=None,
                                   use_fallback_per_query=True,
                                   on_search_complete_hook=None,
                                   **kwargs):
    """Retrieve top ``N`` search results as document instances

    Parameters
    ----------
    queries : collection of str
        Collection of strings representing google queries. Documents for
        the top `num_urls` google search results (from all of these
        queries _combined_ will be returned from this function.
    search_engines : iterable of str
        Ordered collection of search engine names to attempt for web
        search. If the first search engine in the list returns a set
        of URLs, then iteration will end and documents for each URL will
        be returned. Otherwise, the next engine in this list will be
        used to run the web search. If this also fails, the next engine
        is used and so on. If all web searches fail, an empty list is
        returned. See :obj:`~elm.web.search.run.SEARCH_ENGINE_OPTIONS`
        for supported search engine options.
        By default, ``("PlaywrightGoogleLinkSearch",
        "PlaywrightDuckDuckGoLinkSearch",
        "DuxDistributedGlobalSearch")``.
    num_urls : int, optional
        Number of unique top Google search result to return as docs. The
        google search results from all queries are interleaved and the
        top `num_urls` unique URL's are downloaded as docs. If this
        number is less than ``len(queries)``, some of your queries may
        not contribute to the final output. By default, ``None``, which
        sets ``num_urls = 3 * len(queries)``.
    ignore_url_parts : iterable of str, optional
        Optional URL components to blacklist. For example, supplying
        `ignore_url_parts={"wikipedia.org"}` will ignore all URLs that
        contain "wikipedia.org". By default, ``None``.
    search_semaphore : :class:`asyncio.Semaphore`, optional
        Semaphore instance that can be used to limit the number of
        playwright browsers used to submit search engine queries open
        concurrently. For backwards-compatibility, if this input is
        ``None``, the input from `browser_semaphore` will be used in its
        place (i.e. the searches and file downloads will be limited
        using the same semaphore). By default, ``None``.
    browser_semaphore : :class:`asyncio.Semaphore`, optional
        Semaphore instance that can be used to limit the number of
        playwright browsers used to download files open concurrently.
        If ``None``, no limits are applied. By default, ``None``.
    task_name : str, optional
        Optional task name to use in :func:`asyncio.create_task`.
        By default, ``None``.
    use_fallback_per_query : bool, default=True
        Option to use the fallback list of search engines on a per-query
        basis. This means if a single query fails with one search
        engine, the fallback search engines will be attempted for that
        query. If this input is ``False``, the fallback search engines
        are only used if *all* search queries fail for a single search
        engine. By default, ``True``.
    on_search_complete_hook : callable, optional
        If provided, this async callable will be called after the search
        engine links have been retrieved. A single argument will be
        passed to this function containing a list of URL's that were the
        result of the search queries (this list cna be empty if the
        search failed). By default, ``None``.
    **kwargs
        Keyword-argument pairs to initialize
        :class:`elm.web.file_loader.AsyncFileLoader`. This input can
        also include and any/all of the following keywords:

            - ddg_api_kwargs
            - google_cse_api_kwargs
            - google_serper_api_kwargs
            - tavily_api_kwargs
            - ddgs_kwargs
            - cf_google_se_kwargs
            - pw_bing_se_kwargs
            - pw_ddg_se_kwargs
            - pw_google_cse_kwargs
            - pw_google_se_kwargs
            - pw_yahoo_se_kwargs
            - pw_launch_kwargs

        Each of these inputs should be a dictionary with
        keyword-argument pairs that you can use to initialize the search
        engines in the `search_engines` input. If ``pw_launch_kwargs``
        is detected, it will be added to the kwargs for all of the
        PLaywright-based search engines so that you do not have to
        repeatedly specify the launch parameters. For example, you may
        specify ``pw_launch_kwargs={"headless": False}`` to
        have all Playwright-based searches show the browser and _also_
        specify ``google_serper_api_kwargs={"api_key": "..."}`` to
        specify the API key for the Google Serper search.

    Returns
    -------
    list of :class:`elm.web.document.BaseDocument`
        List of documents representing the top `num_urls` results from
        the google searches across all `queries`.
    """
    if search_semaphore is None:
        # backward-compatibility
        search_semaphore = browser_semaphore

    fpq = use_fallback_per_query
    urls = await search_with_fallback(queries, search_engines=search_engines,
                                      num_urls=num_urls,
                                      ignore_url_parts=ignore_url_parts,
                                      browser_semaphore=search_semaphore,
                                      task_name=task_name,
                                      use_fallback_per_query=fpq, **kwargs)
    if on_search_complete_hook is not None:
        await on_search_complete_hook(urls)

    logger.debug("Downloading documents for URLS: \n\t-%s", "\n\t-".join(urls))
    docs = await load_docs(urls, browser_semaphore, **kwargs)
    return docs


async def search_with_fallback(queries, search_engines=_DEFAULT_SE,
                               num_urls=None, ignore_url_parts=None,
                               browser_semaphore=None, task_name=None,
                               use_fallback_per_query=True, **kwargs):
    """Retrieve search query URLs using multiple search engines if needed

    Parameters
    ----------
    queries : collection of str
        Collection of strings representing google queries. Documents for
        the top `num_urls` google search results (from all of these
        queries _combined_ will be returned from this function.
    search_engines : iterable of str
        Ordered collection of search engine names to attempt for web
        search. If the first search engine in the list returns a set
        of URLs, then iteration will end and documents for each URL will
        be returned. Otherwise, the next engine in this list will be
        used to run the web search. If this also fails, the next engine
        is used and so on. If all web searches fail, an empty list is
        returned. See :obj:`~elm.web.search.run.SEARCH_ENGINE_OPTIONS`
        for supported search engine options.
        By default, ``("PlaywrightGoogleLinkSearch", )``.
    num_urls : int, optional
        Number of unique top Google search result to return as docs. The
        google search results from all queries are interleaved and the
        top `num_urls` unique URL's are downloaded as docs. If this
        number is less than ``len(queries)``, some of your queries may
        not contribute to the final output. By default, ``None``, which
        sets ``num_urls = 3 * len(queries)``.
    ignore_url_parts : iterable of str, optional
        Optional URL components to blacklist. For example, supplying
        `ignore_url_parts={"wikipedia.org"}` will ignore all URLs that
        contain "wikipedia.org". By default, ``None``.
    browser_semaphore : :class:`asyncio.Semaphore`, optional
        Semaphore instance that can be used to limit the number of
        playwright browsers open concurrently. If ``None``, no limits
        are applied. By default, ``None``.
    task_name : str, optional
        Optional task name to use in :func:`asyncio.create_task`.
        By default, ``None``.
    use_fallback_per_query : bool, default=True
        Option to use the fallback list of search engines on a per-query
        basis. This means if a single query fails with one search
        engine, the fallback search engines will be attempted for that
        query. If this input is ``False``, the fallback search engines
        are only used if *all* search queries fail for a single search
        engine. By default, ``True``.
    **kwargs
        Keyword-argument pairs to initialize search engines. This input
        can include and any/all of the following keywords:

            - ddg_api_kwargs
            - google_cse_api_kwargs
            - google_serper_api_kwargs
            - tavily_api_kwargs
            - ddgs_kwargs
            - cf_google_se_kwargs
            - pw_bing_se_kwargs
            - pw_ddg_se_kwargs
            - pw_google_cse_kwargs
            - pw_google_se_kwargs
            - pw_yahoo_se_kwargs
            - pw_launch_kwargs

        Each of these inputs should be a dictionary with
        keyword-argument pairs that you can use to initialize the search
        engines in the `search_engines` input. If ``pw_launch_kwargs``
        is detected, it will be added to the kwargs for all of the
        PLaywright-based search engines so that you do not have to
        repeatedly specify the launch parameters. For example, you may
        specify ``pw_launch_kwargs={"headless": False}`` to
        have all Playwright-based searches show the browser and _also_
        specify ``google_serper_api_kwargs={"api_key": "..."}`` to
        specify the API key for the Google Serper search.

    Returns
    -------
    set
        Set of URL's for the queries (empty if search failed).

    Raises
    ------
    ELMInputError
        If `search_engines` input is empty.
    """
    num_urls = num_urls or 3 * len(queries)
    if len(search_engines) < 1:
        msg = f"Must provide at least one search engine! Got {search_engines=}"
        logger.error(msg)
        raise ELMInputError(msg)

    if use_fallback_per_query:
        urls = await _multi_se_search(search_engines, queries, num_urls,
                                      ignore_url_parts, browser_semaphore,
                                      task_name, kwargs)
        if urls:
            return urls
    else:
        for se_name in search_engines:
            logger.debug("Searching web using %r", se_name)
            urls = await _single_se_search(se_name, queries, num_urls,
                                           ignore_url_parts, browser_semaphore,
                                           task_name, kwargs)
            if urls:
                return urls

    logger.warning("No web results found using %d search engines: %r",
                   len(search_engines), search_engines)
    return set()


async def load_docs(urls, browser_semaphore=None, **kwargs):
    """Load a document for each input URL

    Parameters
    ----------
    urls : iterable of str
        Iterable of URL's (as strings) to fetch.
    browser_semaphore : :class:`asyncio.Semaphore`, optional
        Semaphore instance that can be used to limit the number of
        playwright browsers open concurrently for document retrieval. If
        ``None``, no limits are applied. By default, ``None``.
    kwargs
        Keyword-argument pairs to initialize
        :class:`elm.web.file_loader.AsyncFileLoader`.

    Returns
    -------
    list
        List of non-empty document instances containing information from
        the URL's. If a URL could not be fetched (i.e. document instance
        is empty), it will not be included in the output list.
    """
    logger.trace("Downloading docs for the following URL's:\n%r", urls)
    logger.trace("kwargs for AsyncFileLoader:\n%s",
                 pprint.PrettyPrinter().pformat(kwargs))
    file_loader = AsyncFileLoader(browser_semaphore=browser_semaphore,
                                  **kwargs)
    docs = await file_loader.fetch_all(*urls)

    page_lens = {doc.attrs.get("source", "Unknown"): len(doc.pages)
                 for doc in docs}
    logger.debug("Loaded the following number of pages for docs:\n%s",
                 pprint.PrettyPrinter().pformat(page_lens))
    return [doc for doc in docs if not doc.empty]


async def _single_se_search(se_name, queries, num_urls, ignore_url_parts,
                            browser_sem, task_name, kwargs):
    """Search for links using a single search engine"""
    _validate_se_name(se_name)
    links = await _run_search(se_name, queries, browser_sem, task_name, kwargs)
    return _down_select_urls(links, num_urls=num_urls,
                             ignore_url_parts=ignore_url_parts)


async def _multi_se_search(search_engines, queries, num_urls,
                           ignore_url_parts, browser_sem, task_name, kwargs):
    """Search for links using one or more search engines as fallback"""
    outputs = {q: None for q in queries}
    remaining_queries = list(queries)
    for se_name in search_engines:
        _validate_se_name(se_name)

        logger.debug("Searching web using %r", se_name)
        links = await _run_search(se_name, remaining_queries, browser_sem,
                                  task_name, kwargs)
        logger.trace("Links: %r", links)

        failed_queries = []
        for q, se_result in zip(remaining_queries, links):
            if not se_result or not se_result[0]:
                failed_queries.append(q)
                continue
            outputs[q] = se_result

        remaining_queries = failed_queries
        logger.trace("Remaining queries to search: %r", remaining_queries)

        if not remaining_queries:
            break

    links = [link or [[]] for link in outputs.values()]

    return _down_select_urls(links, num_urls=num_urls,
                             ignore_url_parts=ignore_url_parts)


async def _run_search(se_name, queries, browser_sem, task_name, kwargs):
    """Run a search for multiple queries on a single search engine"""
    searchers = [asyncio.create_task(_single_query_search(se_name, query,
                                                          browser_sem, kwargs),
                                     name=task_name) for query in queries]
    return await asyncio.gather(*searchers)


async def _single_query_search(se_name, query, browser_sem, kwargs):
    """Execute a single search query on a single search engine"""
    try:
        search_engine, uses_browser = _init_se(se_name, kwargs)
    except Exception as e:
        logger.error("Could not instantiate %s", se_name)
        logger.exception(e)
        return [[]]

    if uses_browser:
        # help avoid some detection by staggering the browser launches
        await asyncio.sleep(random.uniform(1, 10))
        return await _single_query_pw(search_engine, query,
                                      browser_sem=browser_sem)

    return await _single_query_api(search_engine, query)


async def _single_query_pw(search_engine, question, browser_sem):
    """Perform a single browser-based search"""
    if browser_sem is None:
        browser_sem = AsyncExitStack()

    logger.trace("Single search browser_semaphore=%r", browser_sem)
    async with browser_sem:
        logger.trace("Starting %s search for %r with browser_semaphore=%r",
                     search_engine._SE_NAME, question, browser_sem)
        return await search_engine.results(question,
                                           num_results=_RESULTS_PER_QUERY)


async def _single_query_api(search_engine, question):
    """Perform a single api-based search"""
    logger.trace("Starting %s search for %r", search_engine._SE_NAME, question)
    return await search_engine.results(question,
                                       num_results=_RESULTS_PER_QUERY)


def _init_se(se_name, kwargs):
    """Initialize a search engine class"""
    se_class, uses_browser, kwarg_key = SEARCH_ENGINE_OPTIONS[se_name]
    init_kwargs = {}
    if uses_browser:
        init_kwargs = kwargs.get("pw_launch_kwargs", {})

    # Cannot use 'pop' here because _run_search is invoked once per query.
    init_kwargs.update(kwargs.get(kwarg_key, {}))
    return se_class(**init_kwargs), uses_browser


def _down_select_urls(search_results, num_urls=5, ignore_url_parts=None):
    """Select the top N URLs"""
    ignore_url_parts = _as_set(ignore_url_parts)
    all_urls = chain.from_iterable(zip_longest(*[results[0]
                                                 for results
                                                 in search_results]))
    urls = set()
    for url in all_urls:
        if not url or any(substr in url for substr in ignore_url_parts):
            continue
        urls.add(url)
        if len(urls) == num_urls:
            break
    return urls


def _as_set(user_input):
    """Convert user input (possibly None or str) to set of strings"""
    if isinstance(user_input, str):
        user_input = {user_input}
    return set(user_input or [])


def _validate_se_name(se_name):
    """Validate user search engine name input"""
    if se_name not in SEARCH_ENGINE_OPTIONS:
        msg = (f"'se_name' must be one of: {list(SEARCH_ENGINE_OPTIONS)}\n"
               f"Got {se_name=}")
        logger.error(msg)
        raise ELMKeyError(msg)



# -*- coding: utf-8 -*-
"""ELM Ordinance CPU-bound services"""
import asyncio
from functools import partial
from concurrent.futures import ProcessPoolExecutor

from elm.ords.services.base import Service
from elm.web.document import PDFDocument
from elm.utilities.parse import read_pdf, read_pdf_ocr


class ProcessPoolService(Service):
    """Service that contains a ProcessPoolExecutor instance"""

    def __init__(self, **kwargs):
        """

        Parameters
        ----------
        **kwargs
            Keyword-value argument pairs to pass to
            :class:`concurrent.futures.ProcessPoolExecutor`.
            By default, ``None``.
        """
        self._ppe_kwargs = kwargs or {}
        self.pool = None

    def acquire_resources(self):
        """Open thread pool and temp directory"""
        self.pool = ProcessPoolExecutor(**self._ppe_kwargs)

    def release_resources(self):
        """Shutdown thread pool and cleanup temp directory"""
        self.pool.shutdown(wait=True, cancel_futures=True)


class PDFLoader(ProcessPoolService):
    """Class to load PDFs in a ProcessPoolExecutor."""

    @property
    def can_process(self):
        """bool: Always ``True`` (limiting is handled by asyncio)"""
        return True

    async def process(self, fn, pdf_bytes, **kwargs):
        """Write URL doc to file asynchronously.

        Parameters
        ----------
        doc : elm.web.document.Document
            Document containing meta information about the file. Must
            have a "source" key in the `metadata` dict containing the
            URL, which will be converted to a file name using
            :func:`compute_fn_from_url`.
        file_content : str | bytes
            File content, typically string text for HTML files and bytes
            for PDF file.
        make_name_unique : bool, optional
            Option to make file name unique by adding a UUID at the end
            of the file name. By default, ``False``.

        Returns
        -------
        Path
            Path to output file.
        """
        loop = asyncio.get_running_loop()
        result = await loop.run_in_executor(
            self.pool, partial(fn, pdf_bytes, **kwargs)
        )
        return result


def _read_pdf(pdf_bytes, **kwargs):
    """Utility function so that pdftotext.PDF doesn't have to be pickled."""
    pages = read_pdf(pdf_bytes, verbose=False)
    return PDFDocument(pages, **kwargs)


def _read_pdf_ocr(pdf_bytes, tesseract_cmd, **kwargs):
    """Utility function that mimics `_read_pdf`."""
    if tesseract_cmd:
        _configure_pytesseract(tesseract_cmd)

    pages = read_pdf_ocr(pdf_bytes, verbose=True)
    return PDFDocument(pages, **kwargs)


def _configure_pytesseract(tesseract_cmd):
    """Set the tesseract_cmd"""
    import pytesseract

    pytesseract.pytesseract.tesseract_cmd = tesseract_cmd


async def read_pdf_doc(pdf_bytes, **kwargs):
    """Read PDF file from bytes in a Process Pool.

    Parameters
    ----------
    pdf_bytes : bytes
        Bytes containing PDF file.
    **kwargs
        Keyword-value arguments to pass to
        :class:`elm.web.document.PDFDocument` initializer.

    Returns
    -------
    elm.web.document.PDFDocument
        PDFDocument instances with pages loaded as text.
    """
    return await PDFLoader.call(_read_pdf, pdf_bytes, **kwargs)


async def read_pdf_doc_ocr(pdf_bytes, **kwargs):
    """Read PDF file from bytes using OCR (pytesseract) in a Process Pool.

    Note that Pytesseract must be set up properly for this method to
    work. In particular, the `pytesseract.pytesseract.tesseract_cmd`
    attribute must be set to point to the pytesseract exe.

    Parameters
    ----------
    pdf_bytes : bytes
        Bytes containing PDF file.
    **kwargs
        Keyword-value arguments to pass to
        :class:`elm.web.document.PDFDocument` initializer.

    Returns
    -------
    elm.web.document.PDFDocument
        PDFDocument instances with pages loaded as text.
    """
    import pytesseract

    return await PDFLoader.call(
        _read_pdf_ocr,
        pdf_bytes,
        tesseract_cmd=pytesseract.pytesseract.tesseract_cmd,
        **kwargs
    )



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: disable=no-member
"""ELM Document retrieval from a website"""

import logging
from asyncio import PriorityQueue
from math import inf as infinity
from contextlib import aclosing
from functools import lru_cache

from crawl4ai import AsyncWebCrawler, CrawlerRunConfig, BrowserConfig
from crawl4ai.utils import normalize_url_for_deep_crawl
from crawl4ai.content_scraping_strategy import LXMLWebScrapingStrategy
from crawl4ai.deep_crawling import BestFirstCrawlingStrategy
from crawl4ai.deep_crawling.filters import (FilterChain, URLPatternFilter,
                                            ContentTypeFilter, URLFilter)

from elm.web.file_loader import AsyncFileLoader
from elm.web.document import HTMLDocument


logger = logging.getLogger(__name__)
BEST_ZONING_ORDINANCE_KEYWORDS = {"pdf": 1152,
                                  "zoning": 576,
                                  "ordinance": 288,
                                  "planning": 144,
                                  "plan": 72,
                                  "government": 36,
                                  "code": 12,
                                  "area": 12,
                                  "land development": 6,
                                  "land": 3,
                                  "municipal": 1,
                                  "department": 1}
"""Keywords and their associated point values for scoring URLs.

These keywords are used to prioritize URLs based on their relevance
to contain documents of interest.

The general scoring strategy employed here is that the top keyword
should be worth more points than the sum of all of the keywords below
it. This is just the default strategy, and can/should be modified by
users to better suit their own use cases.
"""

_BLACKLIST_SUBSTRINGS = ["*login*",
                         "*signup*",
                         "*sign up*",
                         r"*sign%20up*",
                         "*signin*",
                         "*sign in*",
                         r"*sign%20in*",
                         "*register*",
                         "*subscribe*",
                         "*donate*",
                         "*shop*",
                         "*cart*",
                         "*careers*",
                         "*events*",
                         "*calendar*"]

ELM_URL_FILTER = URLPatternFilter(reverse=True, patterns=_BLACKLIST_SUBSTRINGS)
"""Filter used to exclude URLs that are not relevant to the search"""

_SCORE_KEY = "website_link_relevance_score"


class PeekablePriorityQueue(PriorityQueue):
    """A priority queue that allows peeking at the next item"""

    def peek(self):
        """Peek at the next item in the queue without removing it"""
        if self.empty():
            return None
        return self._queue[0]


class ELMLinkScorer:
    """Custom URL scorer for ELM website crawling"""

    def __init__(self, keyword_points=None):
        """

        Parameters
        ----------
        keyword_points : dict, optional
            Dictionary mapping keywords to their associated point values
            for scoring URLs. If ``None``, uses
            :obj:`BEST_ZONING_ORDINANCE_KEYWORDS`.
            By default, ``None``.
        """
        self.keyword_points = keyword_points or BEST_ZONING_ORDINANCE_KEYWORDS

    async def score(self, links):
        """Score a link based on its title text and URL.

        The score is calculated by summing the point values of the
        keywords found in both the link title text and the URL. The
        higher the score, the more relevant the link is considered to be
        for the purpose of retrieving zoning ordinance documents.

        Parameters
        ----------
        links : list of dicts
            A list of dictionaries representing links to be scored.
            Each dictionary should contain at least the keys "text" and
            "href". The "text" key should contain the link title text,
            and the "href" key should contain the URL.

        Returns
        -------
        links : list of dicts
            The input list of links with an additional key "score" added
            to each dictionary. Each "score" key contains the calculated
            score for that link.
        """
        for link in links:
            link["score"] = (self._assign_value(link.get("text", ""))
                             + self._assign_value(link.get("href", "")))

        return links

    def _assign_value(self, text):
        """Assign a score based on the presence of keywords in the text"""
        score = 0
        for kw, kw_score in self.keyword_points.items():
            if kw in text.casefold():
                score += kw_score
        return score


class ContentTypeExcludeFilter(URLFilter):
    """Content type to exclude filter using fast lookups"""

    __slots__ = ("exclude_ext", )

    # Fast extension to mime type mapping
    EXCLUDE_EXTENSIONS = {# Images
                          'bmp', 'gif', 'ico', 'jpeg', 'jpg', 'png', 'svg',
                          'tiff', 'webp',
                          # Audio
                          'aac', 'm4a', 'mp3', 'ogg', 'wav',
                          # Video
                          'avi', 'flv', 'mkv', 'mp4', 'mpeg', 'mov', 'webm',
                          'wmv',
                          # Applications
                          '7z', 'exe', 'gz', 'json', 'msi', 'pdf', 'rar',
                          'tar', 'xml', 'zip',
                          # Fonts
                          'otf', 'ttf', 'woff', 'woff2',
                          # OpenDocument Formats
                          'odp', 'ods', 'odt',
                          # Archives
                          'bz2', 'tar.gz', 'tgz',
                          # Others
                          'ai', 'apk', 'bin', 'deb', 'dmg', 'eps', 'epub',
                          'iso', 'jar', 'mid', 'midi', 'ps', 'rpm', 'rtf',
                          'sqlite', 'swf'}
    """File extensions to exclude from the crawl."""

    def __init__(self, exclude_extensions=None):
        """

        Parameters
        ----------
        exclude_extensions : str or list of str, optional
            File extensions to exclude from the crawl. If a string is
            provided, it is treated as a single extension. If a list is
            provided, each item in the list is treated as an extension.
            If ``None``, uses the default set of excluded extensions
            defined in
            :obj:`ContentTypeExcludeFilter.EXCLUDE_EXTENSIONS`.
            By default, ``None``.
        """
        super().__init__()

        if isinstance(exclude_extensions, str):
            exclude_extensions = [exclude_extensions]

        self.exclude_ext = frozenset(t.casefold()
                                     for t in (exclude_extensions
                                               or self.EXCLUDE_EXTENSIONS))

    @lru_cache(maxsize=1000)
    def _check_url_cached(self, url: str) -> bool:
        """Cached URL checking"""

        ext = ContentTypeFilter._extract_extension(url)
        if not ext:
            return True

        return ext not in self.exclude_ext

    def apply(self, url: str) -> bool:
        """Fast extension check with caching"""
        result = self._check_url_cached(url)
        self._update_stats(result)
        return result


class ELMWebsiteCrawlingStrategy(BestFirstCrawlingStrategy):
    """Custom crawling strategy for ELM website searching"""

    BATCH_SIZE = 10
    """Number of URLs to process in each batch"""

    ONE_SCORE_AT_A_TIME = True
    """Whether to batch process only links with the same score.

    This works best if the score is an integer value, since scores are
    compared directly using the `==` operator.
    """

    @classmethod
    async def found_enough_docs(cls, out_docs):
        """Check if enough documents have been found.

        If :obj:`ELMWebsiteCrawlingStrategy.ONE_SCORE_AT_A_TIME` is
        ``True``, this function returns ``True`` when 5 documents have
        been found, otherwise it returns ``True`` when 8 documents
        have been found.

        Parameters
        ----------
        out_docs : list
            List of documents found during the crawl.

        Returns
        -------
        bool
            Whether enough documents have been found to stop the crawl.
        """
        doc_threshold = 5 if cls.ONE_SCORE_AT_A_TIME else 8
        return len(out_docs) >= doc_threshold

    async def link_discovery(self, result, source_url, current_depth, visited,
                             next_links):
        """
        Extract links from the crawl result, validate them, and append new
        URLs (with their parent references) to next_links.
        Also updates the depths dictionary.

        .. NOTE:: Overridden from original BestFirstCrawlingStrategy to
                  implement return full link dictionary instead of just
                  URLs.
        """
        new_depth = current_depth + 1
        if new_depth > self.max_depth:
            return

        remaining_capacity = self.max_pages - self._pages_crawled
        if remaining_capacity <= 0:
            self.logger.info(f"Max pages limit ({self.max_pages}) reached, "
                             "stopping link discovery")
            return

        links = []
        for link in result.links.get("internal", []):
            link["is_external"] = False
            links.append(link)
        if self.include_external:
            for link in result.links.get("external", []):
                link["is_external"] = True
                links.append(link)

        valid_links = []
        for link in links:
            url = link.get("href")
            if not url:
                self.logger.debug("Link without href, skipping")
                continue
            base_url = normalize_url_for_deep_crawl(url, source_url)
            if base_url in visited:
                continue
            if not await self.can_process_url(url, new_depth):
                self.stats.urls_skipped += 1
                continue

            valid_links.append(link)

        for link in valid_links:
            link["source_url"] = source_url
            link["depth"] = new_depth
            next_links.append(link)

    async def _arun_best_first(self, start_url, crawler, config):
        """
        Core best-first crawl method using a priority queue.

        The queue items are tuples of
        (score, depth, url, parent_url, is_external).
        Lower scores are treated as higher priority.
        URLs are processed in batches for efficiency.

        .. NOTE:: Overridden from original BestFirstCrawlingStrategy to
                  implement pass the full link dictionary to the
                  `url_scorer` (instead of just the URL string). Also
                  fixes queue scoring bug in the original code. Also
                  don't allow discovering links from external URLs.

        """
        queue = PeekablePriorityQueue()
        await queue.put((0, 0, start_url, None, False))
        visited = set()

        while not queue.empty() and not self._cancel_event.is_set():
            if self._pages_crawled >= self.max_pages:
                self.logger.info(f"Max pages limit ({self.max_pages}) "
                                 "reached, stopping crawl")
                break

            batch = []
            curr_score = None
            for _ in range(self.BATCH_SIZE):
                if queue.empty():
                    break
                if self.ONE_SCORE_AT_A_TIME and curr_score is not None:
                    item = queue.peek()
                    if item[0] != curr_score:
                        break

                item = await queue.get()
                score, depth, url, parent_url, is_external = item
                if url in visited:
                    continue
                visited.add(url)
                batch.append(item)

                if self.ONE_SCORE_AT_A_TIME and curr_score is None:
                    curr_score = score

            if not batch:
                continue

            urls = [item[2] for item in batch]
            batch_config = config.clone(deep_crawl_strategy=None, stream=True)
            stream_gen = await crawler.arun_many(urls=urls,
                                                 config=batch_config)
            async for result in stream_gen:
                result_url = result.url
                corresponding = next((item
                                      for item in batch
                                      if item[2] == result_url), None)
                if not corresponding:
                    continue

                score, depth, url, parent_url, is_external = corresponding
                result.metadata = result.metadata or {}
                result.metadata["depth"] = depth
                result.metadata["parent_url"] = parent_url
                result.metadata["score"] = -1 * score

                if result.success:
                    self._pages_crawled += 1

                yield result

                if result.success and not is_external:
                    new_links = []
                    await self.link_discovery(result, result_url, depth,
                                              visited, new_links)

                    if self.url_scorer:
                        new_links = await self.url_scorer(new_links)

                    for link in new_links:
                        await queue.put((-1 * link.get("score", 0),
                                         link.get("depth", depth + 1),
                                         link.get("href"),
                                         link.get("source_url"),
                                         link.get("is_external", False)))


class ELMWebsiteCrawler:
    """Crawl a website for documents of interest"""

    def __init__(self, validator, file_loader_kwargs=None,
                 browser_config_kwargs=None, crawl_strategy_kwargs=None,
                 crawler_config_kwargs=None, cte_kwargs=None,
                 extra_url_filters=None, include_external=False,
                 url_scorer=None, max_pages=100, page_limit=None):
        """

        Parameters
        ----------
        validator : callable
            An async callable that takes a document instance (containing
            the text from a PDF or a webpage) and returns a boolean
            indicating whether the text passes the validation check.
            This is used to determine whether or not to keep (i.e.
            return) the document.
        file_loader_kwargs : dict, optional
            Additional keyword-value argument pairs to pass to the
            :class:`~elm.web.file_loader.AsyncFileLoader` class.
            By default, ``None``.
        browser_config_kwargs : dict, optional
            Additional keyword-value argument pairs to pass to the
            :class:`crawl4ai.async_configs.BrowserConfig` class.
            By default, ``None``.
        crawl_strategy_kwargs : dict, optional
            Additional keyword-value argument pairs to pass to the
            :class:`ELMWebsiteCrawlingStrategy` class.
            By default, ``None``.
        crawler_config_kwargs : dict, optional
            Additional keyword-value argument pairs to pass to the
            :class:`crawl4ai.async_configs.CrawlerRunConfig` class.
            By default, ``None``.
        cte_kwargs : dict, optional
            Additional keyword-value argument pairs to pass to the
            :class:`ContentTypeExcludeFilter` class. This filter is used
            to exclude URLs based on their content type.
            By default, ``None``.
        extra_url_filters : list, optional
            Additional URL filters to apply during crawling. Each filter
            must have a (non-async) ``apply`` method that takes a URL
            and returns a boolean indicating whether the URL should be
            included in the crawl. By default, ``None``.
        include_external : bool, optional
            Whether to include external links in the crawl.
            By default, ``False``.
        url_scorer : callable, optional
            An async callable that takes a list of dictionaries
            containing URL information and assigns each dictionary a
            `score` key representing the score for that URL. The input
            URL dictionaries will each have at least one key: "href".
            This key will contain the URL of the link. The dictionary
            may also have other attributes such as "text", which
            contains the link title text. If ``None``, uses the
            :meth:`ELMLinkScorer.score` method to score the URLs.
            By default, ``None``.
        max_pages : int, optional
            Maximum number of **successful** pages to crawl.
            By default, ``100``.
        page_limit : int, optional
            Maximum number of pages to crawl regardless of success
            status. If ``None``, a page limit of 2 * `max_pages` is
            used. To set no limit (not recommended), use ``math.inf``.
            By default, ``None``.
        """
        self.validator = validator
        self.page_limit = page_limit or 2 * max_pages

        flk = {"verify_ssl": False}
        flk.update(file_loader_kwargs or {})
        self.afl = AsyncFileLoader(**flk)

        bck = {"headless": True, "verbose": False}
        bck.update(browser_config_kwargs or {})
        self.browser_config = BrowserConfig(**bck)

        cte_kwargs = cte_kwargs or {}
        url_filters = [ELM_URL_FILTER, ContentTypeExcludeFilter(**cte_kwargs)]
        url_filters += extra_url_filters or []
        self.filter_chain = FilterChain(url_filters)

        strategy_kwargs = {"max_depth": infinity,
                           "include_external": include_external,
                           "filter_chain": self.filter_chain,
                           "url_scorer": url_scorer or ELMLinkScorer().score,
                           "max_pages": max_pages,
                           "logger": logger}
        strategy_kwargs.update(crawl_strategy_kwargs or {})
        self.crawl_strategy = ELMWebsiteCrawlingStrategy(**strategy_kwargs)

        cck = {"deep_crawl_strategy": self.crawl_strategy,
               "scraping_strategy": LXMLWebScrapingStrategy(),
               "stream": True,
               "verbose": False,
               "pdf": False,
               "log_console": False,
               "exclude_social_media_domains": ["youtube.com"],
               "exclude_social_media_links": True,
               "semaphore_count": 1}
        cck.update(crawler_config_kwargs or {})
        self.config = CrawlerRunConfig(**cck)

    async def run(self, base_url, termination_callback=None,
                  on_result_hook=None, return_c4ai_results=False):
        """Crawl a website for documents of interest

        Parameters
        ----------
        base_url : str
            The base URL to start crawling from.
        termination_callback : callable, optional
            An async callable that takes a list of documents and returns
            a boolean indicating whether to stop crawling. If ``None``,
            the :meth:`ELMWebsiteCrawlingStrategy.found_enough_docs` is
            used, which simply terminates when roughly a handful of
            documents have been found. By default, ``None``.
        on_result_hook : callable, optional
            An async callable that is called every time a result is
            found during the crawl. This can be used to perform
            additional processing on each result or to monitor the crawl
            progress. The callable should accept a single argument,
            which is the crawl result object. If ``None``, no additional
            processing is done on the results. By default, ``None``.
        return_c4ai_results : bool, optional
            Whether to return the raw crawl4ai results along with the
            documents. If ``True``, returns a tuple of (documents,
            crawl4ai_results). If ``False``, returns only the documents.
            By default, ``False``.

        Returns
        -------
        out_docs
            List of document instances that passed the validation
            check. Each document contains the text from a PDF or a
            webpage, and has an attribute `source` that contains the
            URL of the document.
        results : list, optional
            List of crawl4ai results containing metadata about the
            crawled pages. This is only returned if
            `return_c4ai_results` is ``True``.
        """

        results = []
        out_docs = []
        should_stop = (termination_callback
                       or ELMWebsiteCrawlingStrategy.found_enough_docs)
        page_count = 0
        async with AsyncWebCrawler(config=self.browser_config) as crawler:
            crawl_results = await crawler.arun(base_url, config=self.config)
            async with aclosing(crawl_results) as agen:
                async for result in agen:
                    page_count += 1
                    if page_count > self.page_limit:
                        logger.debug("Exiting crawl due to page limit")
                        break
                    if not result.success:
                        continue
                    results.append(result)
                    logger.debug("Crawled %s", result.url)
                    if on_result_hook:
                        await on_result_hook(result)

                    score = result.metadata.get("score", 0)
                    depth = result.metadata.get("depth", 0)
                    logger.trace("\t- Depth: %d | Score: %.2f", depth, score)

                    doc = await self._doc_from_result(result)
                    doc.attrs[_SCORE_KEY] = score
                    if doc.empty:
                        logger.debug("Empty document, skipping")
                        continue

                    if await self.validator(doc):
                        logger.debug("Document passed validation check")
                        out_docs.append(doc)

                    if await should_stop(out_docs):
                        logger.debug("Exiting crawl early")
                        break

        logger.info("Crawled %d pages", len(results))
        logger.info("Found %d potential documents", len(out_docs))
        logger.debug("Average score: %.2f", _compute_avg_score(results))

        depth_counts = {}
        for result in results:
            depth = result.metadata.get("depth", 0)
            depth_counts[depth] = depth_counts.get(depth, 0) + 1

        logger.debug("Pages crawled by depth:")
        for depth, count in sorted(depth_counts.items()):
            logger.debug(f"  Depth {depth}: {count} pages")

        if out_docs:
            out_docs.sort(key=lambda x: -1 * x.attrs[_SCORE_KEY])

        if return_c4ai_results:
            return out_docs, results

        return out_docs

    async def _doc_from_result(self, result):
        """Get document instance from crawling result"""
        if result.markdown and len(md := result.markdown.strip()) > 3:
            doc = HTMLDocument([md])
            doc.attrs["source"] = result.url
            return doc

        return await self.afl.fetch(result.url)


def _compute_avg_score(results):
    """Compute the average score of the crawled results"""
    if len(results) <= 0:
        return 0
    return sum(r.metadata.get('score', 0) for r in results) / len(results)



# -*- coding: utf-8 -*-
"""ELM Ordinance mutually-exclusive features class."""


class SetbackFeatures:
    """Utility class to get mutually-exclusive feature descriptions."""

    DEFAULT_FEATURE_DESCRIPTIONS = {
        "struct": [
            "occupied dwellings",
            "buildings",
            "structures",
            "residences",
        ],
        "pline": ["property lines", "parcels", "subdivisions"],
        "roads": ["roads"],  # , "rights-of-way"],
        "rail": ["railroads"],
        "trans": [
            "overhead electrical transmission lines",
            "overhead utility lines",
            "utility easements",
            "utility lines",
            "power lines",
            "electrical lines",
            "transmission lines",
        ],
        "water": ["lakes", "reservoirs", "streams", "rivers", "wetlands"],
    }
    FEATURES_AS_IGNORE = {
        "struct": "structures",
        "pline": "property lines",
        "roads": "roads",
        "rail": "railroads",
        "trans": "transmission lines",
        "water": "wetlands",
    }
    FEATURE_CLARIFICATIONS = {
        "struct": "",
        "pline": "",
        "roads": "Roads may also be labeled as rights-of-way. ",
        "rail": "",
        "trans": "",
        "water": "",
    }

    def __init__(self):
        self._validate_descriptions()

    def __iter__(self):
        for feature_id in self.DEFAULT_FEATURE_DESCRIPTIONS:
            feature, ignore = self._keep_and_ignore(feature_id)
            clarification = self.FEATURE_CLARIFICATIONS.get(feature_id, "")
            yield {
                "feature_id": feature_id,
                "feature": feature,
                "ignore_features": ignore,
                "feature_clarifications": clarification,
            }

    def _validate_descriptions(self):
        """Ensure all features have at least one description."""
        features_missing_descriptors = set()
        for feature, descriptions in self.DEFAULT_FEATURE_DESCRIPTIONS.items():
            if len(descriptions) < 1:
                features_missing_descriptors.add(feature)

        if any(features_missing_descriptors):
            raise ValueError(
                f"The following features are missing descriptors: "
                f"{features_missing_descriptors}"
            )

    def _keep_and_ignore(self, feature_id):
        """Get the keep and ignore phrases for a feature."""
        keep_keywords = self.DEFAULT_FEATURE_DESCRIPTIONS[feature_id]
        ignore = [
            keyword
            for feat_id, keyword in self.FEATURES_AS_IGNORE.items()
            if feat_id != feature_id
        ]

        keep_phrase = _join_keywords(keep_keywords, final_sep=", and/or ")
        ignore_phrase = _join_keywords(ignore, final_sep=", and ")

        return keep_phrase, ignore_phrase


def _join_keywords(keywords, final_sep):
    """Join a list of keywords/descriptions."""
    if len(keywords) < 1:
        return ""

    if len(keywords) == 1:
        return keywords[0]

    comma_separated = ", ".join(keywords[:-1])
    return final_sep.join([comma_separated, keywords[-1]])



# -*- coding: utf-8 -*-
"""ELM Web Scraping - Tavily API search"""
import json
import logging
import requests

from tavily import TavilyClient
from tavily.errors import (UsageLimitExceededError, InvalidAPIKeyError,
                           BadRequestError, ForbiddenError)

from elm.web.search.base import APISearchEngineLinkSearch


logger = logging.getLogger(__name__)


class _PatchedTavilyClient(TavilyClient):
    """Patch `TavilyClient` to accept verify keyword"""

    def __init__(self, api_key=None, proxies=None, verify=False):
        """

        Parameters
        ----------
        api_key : str, optional
            API key for search engine. If ``None``, will look up the API
            key using the ``"TAVILY_API_KEY"`` environment variable.
            By default, ``None``.
        verify : bool, default=False
            Option to use SSL verification when making request to API
            endpoint. By default, ``False``.
        """
        super().__init__(api_key=api_key, proxies=proxies)
        self.verify = verify

    def _search(self, query, search_depth="basic", topic="general",
                time_range=None, days=7, max_results=5, include_domains=None,
                exclude_domains=None, include_answer=False,
                include_raw_content=False, include_images=False, timeout=60,
                **kwargs):
        """Internal search method to send the request to the API"""

        data = {"query": query, "search_depth": search_depth, "topic": topic,
                "time_range": time_range, "days": days,
                "include_answer": include_answer,
                "include_raw_content": include_raw_content,
                "max_results": max_results, "include_domains": include_domains,
                "exclude_domains": exclude_domains,
                "include_images": include_images}

        if kwargs:
            data.update(kwargs)

        timeout = min(timeout, 120)

        response = requests.post(self.base_url + "/search",
                                 data=json.dumps(data), headers=self.headers,
                                 timeout=timeout, proxies=self.proxies,
                                 verify=self.verify)

        if response.status_code == 200:
            return response.json()
        else:
            detail = ""
            try:
                detail = response.json().get("detail", {}).get("error", None)
            except Exception:
                pass

            if response.status_code == 429:
                raise UsageLimitExceededError(detail)
            elif response.status_code in [403,432,433]:
                raise ForbiddenError(detail)
            elif response.status_code == 401:
                raise InvalidAPIKeyError(detail)
            elif response.status_code == 400:
                raise BadRequestError(detail)
            else:
                raise response.raise_for_status()

    def _extract(self, urls, include_images=False, extract_depth="basic",
                 timeout=60, **kwargs):
        """
        Internal extract method to send the request to the API.
        """
        data = {"urls": urls, "include_images": include_images,
                "extract_depth": extract_depth}
        if kwargs:
            data.update(kwargs)

        timeout = min(timeout, 120)

        response = requests.post(self.base_url + "/extract",
                                 data=json.dumps(data), headers=self.headers,
                                 timeout=timeout, proxies=self.proxies,
                                 verify=self.verify)

        if response.status_code == 200:
            return response.json()
        else:
            detail = ""
            try:
                detail = response.json().get("detail", {}).get("error", None)
            except Exception:
                pass

            if response.status_code == 429:
                raise UsageLimitExceededError(detail)
            elif response.status_code in [403,432,433]:
                raise ForbiddenError(detail)
            elif response.status_code == 401:
                raise InvalidAPIKeyError(detail)
            elif response.status_code == 400:
                raise BadRequestError(detail)
            else:
                raise response.raise_for_status()


class APITavilySearch(APISearchEngineLinkSearch):
    """Search the web for links using the Tavily API"""

    _SE_NAME = "Tavily API"

    API_KEY_VAR = "TAVILY_API_KEY"
    """Environment variable that should contain the Tavily API key"""

    def __init__(self, api_key=None, verify=False):
        """

        Parameters
        ----------
        api_key : str, optional
            API key for serper search API. If ``None``, will look up the
            API key using the ``"TAVILY_API_KEY"`` environment variable.
            By default, ``None``.
        verify : bool, default=False
            Option to use SSL verification when making request to API
            endpoint. By default, ``False``.
        """
        super().__init__(api_key=api_key)
        self.verify = verify

    async def _search(self, query, num_results=10):
        """Search web for links related to a query"""

        client = _PatchedTavilyClient(api_key=self.api_key, verify=self.verify)
        response = client.search(query=query, max_results=num_results)
        results = response.get("results", [])
        return list(filter(None, (info.get('url', "").replace("+", "%20")
                                  for info in results)))



# -*- coding: utf-8 -*-
"""ELM Ordinance county file downloading logic"""
import logging

from elm.ords.llm import StructuredLLMCaller
from elm.ords.extraction import check_for_ordinance_info
from elm.ords.services.threaded import TempFileCache
from elm.ords.validation.location import CountyValidator
from elm.web.document import PDFDocument
from elm.web.search import web_search_links_as_docs
from elm.web.utilities import filter_documents


logger = logging.getLogger(__name__)
QUESTION_TEMPLATES = [
    '0. "wind energy conversion system zoning ordinances {location}"',
    '1. "{location} wind WECS zoning ordinance"',
    '2. "Where can I find the legal text for commercial wind energy '
    'conversion system zoning ordinances in {location}?"',
    '3. "What is the specific legal information regarding zoning '
    'ordinances for commercial wind energy conversion systems in {location}?"',
]


async def download_county_ordinance(
    location,
    text_splitter,
    num_urls=5,
    file_loader_kwargs=None,
    browser_semaphore=None,
    **kwargs
):
    """Download the ordinance document(s) for a single county.

    Parameters
    ----------
    location : :class:`elm.ords.utilities.location.Location`
        Location objects representing the county.
    text_splitter : obj, optional
        Instance of an object that implements a `split_text` method.
        The method should take text as input (str) and return a list
        of text chunks. Raw text from HTML pages will be passed through
        this splitter to split the single wep page into multiple pages
        for the output document. Langchain's text splitters should work
        for this input.
    num_urls : int, optional
        Number of unique Google search result URL's to check for
        ordinance document. By default, ``5``.
    file_loader_kwargs : dict, optional
        Dictionary of keyword-argument pairs to initialize
        :class:`elm.web.file_loader.AsyncFileLoader` with. If found, the
        "pw_launch_kwargs" key in these will also be used to initialize
        the :class:`elm.web.google_search.PlaywrightGoogleLinkSearch`
        used for the google URL search. By default, ``None``.
    browser_semaphore : :class:`asyncio.Semaphore`, optional
        Semaphore instance that can be used to limit the number of
        playwright browsers open concurrently. If ``None``, no limits
        are applied. By default, ``None``.
    **kwargs
        Keyword-value pairs used to initialize an
        `elm.ords.llm.LLMCaller` instance.

    Returns
    -------
    elm.web.document.BaseDocument | None
        Document instance for the downloaded document, or ``None`` if no
        document was found.
    """
    docs = await _docs_from_google_search(
        location,
        text_splitter,
        num_urls,
        browser_semaphore,
        **(file_loader_kwargs or {})
    )
    docs = await _down_select_docs_correct_location(
        docs, location=location, **kwargs
    )
    docs = await _down_select_docs_correct_content(
        docs, location=location, text_splitter=text_splitter, **kwargs
    )
    logger.info(
        "Found %d potential ordinance documents for %s",
        len(docs),
        location.full_name,
    )
    return _sort_final_ord_docs(docs)


async def _docs_from_google_search(
    location, text_splitter, num_urls, browser_semaphore, **file_loader_kwargs
):
    """Download docs from google location queries. """
    queries = [
        question.format(location=location.full_name)
        for question in QUESTION_TEMPLATES
    ]
    file_loader_kwargs.update(
        {
            "html_read_kwargs": {"text_splitter": text_splitter},
            "file_cache_coroutine": TempFileCache.call,
        }
    )
    return await web_search_links_as_docs(
        queries,
        num_urls=num_urls,
        browser_semaphore=browser_semaphore,
        task_name=location.full_name,
        **file_loader_kwargs,
    )


async def _down_select_docs_correct_location(docs, location, **kwargs):
    """Remove all documents not pertaining to the location."""
    llm_caller = StructuredLLMCaller(**kwargs)
    county_validator = CountyValidator(llm_caller)
    return await filter_documents(
        docs,
        validation_coroutine=county_validator.check,
        task_name=location.full_name,
        county=location.name,
        state=location.state,
    )


async def _down_select_docs_correct_content(docs, location, **kwargs):
    """Remove all documents that don't contain ordinance info."""
    return await filter_documents(
        docs,
        validation_coroutine=_contains_ords,
        task_name=location.full_name,
        **kwargs,
    )


async def _contains_ords(doc, **kwargs):
    """Helper coroutine that checks for ordinance info. """
    doc = await check_for_ordinance_info(doc, **kwargs)
    return doc.attrs.get("contains_ord_info", False)


def _sort_final_ord_docs(all_ord_docs):
    """Sort the final list of documents by year, type, and text length."""
    if not all_ord_docs:
        return None

    return sorted(all_ord_docs, key=_ord_doc_sorting_key)[-1]


def _ord_doc_sorting_key(doc):
    """All text sorting key"""
    year, month, day = doc.attrs.get("date", (-1, -1, -1))
    return year, isinstance(doc, PDFDocument), -1 * len(doc.text), month, day



# -*- coding: utf-8 -*-
"""ELM Ordinance async decision tree."""
import networkx as nx
import logging

from elm.tree import DecisionTree


logger = logging.getLogger(__name__)


class AsyncDecisionTree(DecisionTree):
    """Async class to traverse a directed graph of LLM prompts. Nodes are
    prompts and edges are transitions between prompts based on conditions
    being met in the LLM response


    Purpose:
        Represent a series of prompts that can be used in sequence to
        extract values of interest from text.
    Responsibilities:
        1. Store all prompts used to extract a particular ordinance
           value from text.
        2. Track relationships between the prompts (i.e. which prompts
           is used first, which prompt is used next depending on the
           output of the previous prompt, etc.) using a directed acyclic
           graph.
    Key Relationships:
        Inherits from :class:`~elm.tree.DecisionTree` to add ``async``
        capabilities. Uses a :class:`~elm.ords.llm.calling.ChatLLMCaller`
        for LLm queries.

    .. end desc
    """

    def __init__(self, graph):
        """Async class to traverse a directed graph of LLM prompts. Nodes are
        prompts and edges are transitions between prompts based on conditions
        being met in the LLM response.

        Parameters
        ----------
        graph : nx.DiGraph
            Directed acyclic graph where nodes are LLM prompts and edges are
            logical transitions based on the response. Must have high-level
            graph attribute "chat_llm_caller" which is a ChatLLMCaller
            instance. Nodes should have attribute "prompt" which can have
            {format} named arguments that will be filled from the high-level
            graph attributes. Edges can have attribute "condition" that is a
            callable to be executed on the LLM response text. An edge from a
            node without a condition acts as an "else" statement if no other
            edge conditions are satisfied. A single edge from node to node
            does not need a condition.
        """
        self._g = graph
        self._history = []
        assert isinstance(self.graph, nx.DiGraph)
        assert "chat_llm_caller" in self.graph.graph

    @property
    def chat_llm_caller(self):
        """elm.ords.llm.ChatLLMCaller: ChatLLMCaller instance for this tree."""
        return self.graph.graph["chat_llm_caller"]

    @property
    def messages(self):
        """Get a list of the conversation messages with the LLM.

        Returns
        -------
        list
        """
        return self.chat_llm_caller.messages

    @property
    def all_messages_txt(self):
        """Get a printout of the full conversation with the LLM

        Returns
        -------
        str
        """
        messages = [
            f"{msg['role'].upper()}: {msg['content']}" for msg in self.messages
        ]
        messages = "\n\n".join(messages)
        return messages

    async def async_call_node(self, node0):
        """Call the LLM with the prompt from the input node and search the
        successor edges for a valid transition condition

        Parameters
        ----------
        node0 : str
            Name of node being executed.

        Returns
        -------
        out : str
            Next node or LLM response if at a leaf node.
        """
        prompt = self._prepare_graph_call(node0)
        out = await self.chat_llm_caller.call(prompt, usage_sub_label="dtree")
        logger.debug(
            "Chat GPT prompt:\n%s\nChat GPT response:\n%s", prompt, out
        )
        return self._parse_graph_output(node0, out)

    async def async_run(self, node0="init"):
        """Traverse the decision tree starting at the input node.

        Parameters
        ----------
        node0 : str
            Name of starting node in the graph. This is typically called "init"

        Returns
        -------
        out : str
            Final response from LLM at the leaf node.
        """

        self._history = []

        while True:
            try:
                out = await self.async_call_node(node0)
            except Exception as e:
                logger.debug(
                    "Error traversing trees, here's the full "
                    "conversation printout:\n%s",
                    self.all_messages_txt,
                )
                last_message = self.messages[-1]["content"]
                msg = (
                    "Ran into an exception when traversing tree. "
                    "Last message from LLM is printed below. "
                    "See debug logs for more detail. "
                    "\nLast message: \n"
                    '"""\n%s\n"""'
                )
                logger.error(msg, last_message)
                logger.exception(e)
                raise RuntimeError(msg % last_message) from e
            if out in self.graph:
                node0 = out
            else:
                break

        logger.info("Output: %s", out)

        return out



# -*- coding: utf-8 -*-
"""
Utility to break text up into overlapping chunks.
"""
import copy
from elm.base import ApiBase


class Chunker(ApiBase):
    """
    Class to break text up into overlapping chunks

    NOTE: very large paragraphs that exceed the tokens per chunk will not be
    split up and will still be padded with overlap.
    """

    def __init__(self, text, tag=None, tokens_per_chunk=500, overlap=1,
                 split_on='\n\n'):
        """
        Parameters
        ----------
        text : str
            Single body of text to break up. Works well if this is a single
            document with empty lines between paragraphs.
        tag : None | str
            Optional reference tag to include at the beginning of each text
            chunk
        tokens_per_chunk : float
            Nominal token count per text chunk. Overlap paragraphs will exceed
            this.
        overlap : int
            Number of paragraphs to overlap between chunks
        split_on : str
            Sub string to split text into paragraphs.
        """

        super().__init__()

        self._split_on = split_on
        self._idc = 0  # iter index for chunk
        self.text = self.clean_paragraphs(text)
        self.tag = tag
        self.tokens_per_chunk = tokens_per_chunk
        self.overlap = overlap
        self._paragraphs = None
        self._ptokens = None
        self._ctokens = None
        self._chunks = self.chunk_text()

    def __getitem__(self, i):
        """Get a chunk index

        Returns
        -------
        str
        """
        return self.chunks[i]

    def __iter__(self):
        self._idc = 0
        return self

    def __next__(self):
        """Iterator returns one of the text chunks at a time

        Returns
        -------
        str
        """

        if self._idc >= len(self):
            raise StopIteration

        out = self.chunks[self._idc]
        self._idc += 1
        return out

    def __len__(self):
        """Number of text chunks

        Return
        ------
        int
        """
        return len(self.chunks)

    @property
    def chunks(self):
        """List of overlapping text chunks (strings).

        Returns
        -------
        list
        """
        return self._chunks

    @property
    def paragraphs(self):
        """Get a list of paragraphs in the text demarkated by an empty line.

        Returns
        -------
        list
        """
        if self._paragraphs is None:
            self._paragraphs = self.text.split(self._split_on)
            self._paragraphs = [p for p in self._paragraphs
                                if self.is_good_paragraph(p)]
        return self._paragraphs

    @staticmethod
    def clean_paragraphs(text):
        """Clean up double line breaks to make sure paragraphs can be detected
        in the text."""
        previous_len = len(text)
        while True:
            text = text.replace('\n ', '\n')
            if len(text) == previous_len:
                break
            else:
                previous_len = len(text)
        return text

    @staticmethod
    def is_good_paragraph(paragraph):
        """Basic tests to make sure the paragraph is useful text."""
        if '.....' in paragraph:
            return False
        elif paragraph.strip().isnumeric():
            return False
        else:
            return True

    @property
    def paragraph_tokens(self):
        """Number of tokens per paragraph.

        Returns
        -------
        list
        """
        if self._ptokens is None:
            self._ptokens = [self.count_tokens(p, self.model)
                             for p in self.paragraphs]
        return self._ptokens

    @property
    def chunk_tokens(self):
        """Number of tokens per chunk.

        Returns
        -------
        list
        """
        if self._ctokens is None:
            self._ctokens = [self.count_tokens(c, self.model)
                             for c in self.chunks]
        return self._ctokens

    def merge_chunks(self, chunks_input):
        """Merge chunks until they reach the token limit per chunk.

        Parameters
        ----------
        chunks_input : list
            List of list of integers: [[0, 1], [2], [3, 4]] where nested lists
            are chunks and the integers are paragraph indices

        Returns
        -------
        chunks : list
            List of list of integers: [[0, 1], [2], [3, 4]] where nested lists
            are chunks and the integers are paragraph indices
        """

        chunks = copy.deepcopy(chunks_input)

        for i in range(len(chunks) - 1):
            chunk0 = chunks[i]
            chunk1 = chunks[i + 1]
            if chunk0 is not None and chunk1 is not None:
                tcount0 = sum(self.paragraph_tokens[j] for j in chunk0)
                tcount1 = sum(self.paragraph_tokens[j] for j in chunk1)
                if tcount0 + tcount1 < self.tokens_per_chunk:
                    chunk0 += chunk1
                    chunks[i] = chunk0
                    chunks[i + 1] = None

        chunks = [c for c in chunks if c is not None]
        flat_chunks = [a for b in chunks for a in b]

        assert all(c in list(range(len(self.paragraphs))) for c in flat_chunks)

        return chunks

    def add_overlap(self, chunks_input):
        """Add overlap on either side of a text chunk. This ignores token
        limit.

        Parameters
        ----------
        chunks_input : list
            List of list of integers: [[0, 1], [2], [3, 4]] where nested lists
            are chunks and the integers are paragraph indices

        Returns
        -------
        chunks : list
            List of list of integers: [[0, 1], [2], [3, 4]] where nested lists
            are chunks and the integers are paragraph indices
        """

        if len(chunks_input) == 1 or self.overlap == 0:
            return chunks_input

        chunks = copy.deepcopy(chunks_input)

        for i, chunk1 in enumerate(chunks_input):

            if i == 0:
                chunk2 = chunks_input[i + 1]
                chunk1 = chunk1 + chunk2[:self.overlap]

            elif i == len(chunks) - 1:
                chunk0 = chunks_input[i - 1]
                chunk1 = chunk0[-self.overlap:] + chunk1

            else:
                chunk0 = chunks_input[i - 1]
                chunk2 = chunks_input[i + 1]
                chunk1 = (chunk0[-self.overlap:]
                          + chunk1
                          + chunk2[:self.overlap])

            chunks[i] = chunk1

        return chunks

    def chunk_text(self):
        """Perform the text chunking operation

        Returns
        -------
        chunks : list
            List of strings where each string is an overlapping chunk of text
        """

        chunks_input = [[i] for i in range(len(self.paragraphs))]
        while True:
            chunks = self.merge_chunks(chunks_input)
            if chunks == chunks_input:
                break
            else:
                chunks_input = copy.deepcopy(chunks)

        chunks = self.add_overlap(chunks)
        text_chunks = []
        for chunk in chunks:
            paragraphs = [self.paragraphs[c] for c in chunk]
            text_chunks.append(self._split_on.join(paragraphs))

        if self.tag is not None:
            text_chunks = [self.tag + '\n\n' + chunk for chunk in text_chunks]

        return text_chunks



# -*- coding: utf-8 -*-
"""ELM Web Scraping - Google search."""
import os
import json
import random
import asyncio
import logging
import requests
from contextlib import asynccontextmanager

from camoufox.async_api import AsyncCamoufox
from googleapiclient.discovery import build
from playwright.async_api import TimeoutError as PlaywrightTimeoutError

from elm.web.search.base import (PlaywrightSearchEngineLinkSearch,
                                 APISearchEngineLinkSearch,
                                 PatchedSerpApiClient)


logger = logging.getLogger(__name__)


class PlaywrightGoogleLinkSearch(PlaywrightSearchEngineLinkSearch):
    """Search for top links on the main google search engine

    Purpose:
        Search Google using Playwright engine.
    Responsibilities:
        1. Launch browser using Playwright and
        navigate to Google.
        2. Submit a query to Google Search.
        3. Get list of resulting URLs.
    Key Relationships:
        Relies on `Playwright <https://playwright.dev/python/>`_ for
        web access.

    .. end desc
    """
    MAX_RESULTS_CONSIDERED_PER_PAGE = 10
    """Number of results considered per Google page.

    This value used to be 10, but the addition of extra divs like a set
    of youtube links has brought this number down.
    """

    _SE_NAME = "Google"
    _SE_URL = "https://www.google.com"
    _SE_SR_TAG = '[jsname="UWckNb"]'
    _SC = None
    _SE_QUERY_URL = (
        "https://www.google.com/search?q={}&gl=sg&hl=en&udm=14&start=0&num=10"
    )

    async def _perform_homepage_search(self, page, search_query):
        """Fill in search bar with user query and hit enter"""
        await self._move_mouse(page)

        logger.trace("Clicking on search bar for query: %r", search_query)
        await self._click_on_search_bar(page)

        logger.trace("Filling in search bar for query: %r", search_query)
        await self._fill_in_search_bar(page, search_query)

        logger.trace("Hitting enter for query: %r", search_query)
        await page.keyboard.press('Enter')

    async def _click_on_search_bar(self, page):
        """Find the search bar and click it"""
        try:
            search_bar = page.get_by_label("Search", exact=True)
            return await self._move_and_click(page, search_bar)
        except PlaywrightTimeoutError:
            pass

        search_bar = page.locator('[name="q"]')
        try:
            return self._move_and_click(page, search_bar)
        except PlaywrightTimeoutError:
            pass

        search_bar = page.locator('[autofocus]')
        return self._move_and_click(page, search_bar)

    async def _fill_in_search_bar(self, page, search_query):
        """Attempt to find and fill the search bar several ways"""
        await asyncio.sleep(random.uniform(0.5, 1.5))

        logger.trace("Typing in query: %r", search_query)
        await page.keyboard.type(search_query, delay=random.randint(80, 150))
        return await asyncio.sleep(random.uniform(0.5, 1.5))


class CamoufoxGoogleLinkSearch(PlaywrightGoogleLinkSearch):
    """Search for top links on Google using Camoufox browser"""

    def __init__(self, use_homepage=True, **launch_kwargs):
        """

        Parameters
        ----------
        use_homepage : bool, default=True
            If ``True``, the browser will be navigated to the search
            engine homepage and the query will be input into the search
            bar. If ``False``, the query will be embedded in the URL
            and the browser will navigate directly to the filled-out
            URL. By default, ``False``.
        **launch_kwargs
            Keyword arguments to be passed to
            `playwright.firefox.launch`. For example, you can pass
            ``headless=False, slow_mo=50`` for a visualization of the
            search.
        """
        self.use_homepage = use_homepage
        self.launch_kwargs = {"humanize": 0.1, "headless": True}
        self.launch_kwargs.update(launch_kwargs)
        self._browser = None

    async def _load_browser(self, pw_instance):
        """Empty implementation since we are using camoufox"""

    @asynccontextmanager
    async def _browser_page(self):
        """Get page to use for search"""
        async with AsyncCamoufox(**self.launch_kwargs) as browser:
            yield await browser.new_page()


class PlaywrightGoogleCSELinkSearch(PlaywrightSearchEngineLinkSearch):
    """Search for top links on a custom google search engine

    Resources
    ---------
    https://programmablesearchengine.google.com/controlpanel/create
    """

    _SE_NAME = "Google CSE"
    _SE_SR_TAG = "a.gs-title[href]"
    PAGE_LOAD_TIMEOUT = 10_000
    _SE_QUERY_URL = None

    def __init__(self, cse_url, use_homepage=False, **launch_kwargs):
        """

        Parameters
        ----------
        cse_url : str
            URL of the custom google programmable search engine.
        use_homepage : bool, default=False
            If ``True``, the browser will be navigated to the search
            engine homepage and the query will be input into the search
            bar. If ``False``, the query will be embedded in the URL
            and the browser will navigate directly to the filled-out
            URL. By default, ``False``.
        **launch_kwargs
            Keyword arguments to be passed to
            `playwright.chromium.launch`. For example, you can pass
            ``headless=False, slow_mo=50`` for a visualization of the
            search.
        """
        super().__init__(use_homepage=use_homepage, **launch_kwargs)
        self._cse_url = cse_url
        self._SE_QUERY_URL = (
            f"https://cse.google.com/cse?cx={cse_url}#gsc.tab=0"
            "&gsc.q={}&gsc.sort="
        )

    @property
    def _SE_URL(self):
        """str: URL for the custom google programmable search engine"""
        if not self._cse_url.endswith("#gsc.tab=0"):
            self._cse_url = f"{self._cse_url}#gsc.tab=0"
        return self._cse_url

    async def _perform_homepage_search(self, page, search_query):
        """Fill in search bar with user query and hit enter"""
        logger.trace("Finding search bar for query: %r", search_query)
        await page.get_by_label("search", exact=True).fill(search_query)
        logger.trace("Hitting enter for query: %r", search_query)
        await page.keyboard.press('Enter')

    async def _extract_links(self, page, num_results, query):
        """Extract links for top `num_results` on page"""
        await page.wait_for_load_state("networkidle",
                                       timeout=self.PAGE_LOAD_TIMEOUT)
        await page.wait_for_selector(self._SE_SR_TAG)
        locator = page.locator(self._SE_SR_TAG)

        count = await locator.count() // 2
        links = []

        for i in range(count):
            element = locator.nth(i * 2)
            try:
                link = await element.get_attribute("href")
                if link is not None:
                    links.append(link)
            except Exception:
                logger.exception("Skipped extracting link %d for query %r",
                                 i, query)

            if len(links) >= num_results:
                break

        return links


class APIGoogleCSESearch(APISearchEngineLinkSearch):
    """Search the web for links using a Google CSE API"""

    _BUILD_ARGS = {"serviceName": "customsearch", "version": "v1"}
    _SE_NAME = "Google CSE API"

    API_KEY_VAR = "GOOGLE_API_KEY"
    """Environment variable that should contain the Google CSE API key"""
    CSE_ID_VAR = "GOOGLE_CSE_ID"
    """Environment variable that should contain CSE ID"""

    def __init__(self, api_key=None, cse_id=None):
        """

        Parameters
        ----------
        api_key : str, optional
            API key for search engine. If ``None``, will look up the API
            key using the :obj:`API_KEY_VAR` environment variable.
            By default, ``None``.
        """
        super().__init__(api_key=api_key)
        self.cse_id = cse_id or os.environ.get(self.CSE_ID_VAR or "")

    async def _search(self, query, num_results=10):
        """Search web for links related to a query"""
        build_args = dict(self._BUILD_ARGS)
        build_args["developerKey"] = self.api_key

        search_args = {"q": query, "cx": self.cse_id, "num": num_results}

        results = build(**build_args).cse().list(**search_args).execute()
        results = (results or {}).get('items', [])
        return list(filter(None, (info.get("link") for info in results)))


class SerpAPIGoogleSearch(APISearchEngineLinkSearch):
    """Search the google for links using the SerpAPI service"""

    _SE_NAME = "SerpAPI (Google)"

    API_KEY_VAR = "SERPAPI_KEY"
    """Environment variable that should contain the SerpAPI key"""

    def __init__(self, api_key=None, verify=False):
        """

        Parameters
        ----------
        api_key : str, optional
            API key for serper search API. If ``None``, will look up the
            API key using the ``"SERPAPI_KEY"`` environment variable.
            By default, ``None``.
        verify : bool, default=False
            Option to use SSL verification when making request to API
            endpoint. By default, ``False``.
        """
        super().__init__(api_key=api_key)
        self.verify = verify

    async def _search(self, query, num_results=10, **param_kwargs):
        """Search web for links related to a query"""

        params = {"q": query, "hl": "en", "gl": "us", "api_key": self.api_key}
        params.update(param_kwargs)

        client = PatchedSerpApiClient(params, engine="google",
                                      verify=self.verify)
        results = client.get_dict()
        results = results.get("organic_results", [])
        return list(filter(None, (info.get('link', "").replace("+", "%20")
                                  for info in results)))[:num_results]


class APISerperSearch(APISearchEngineLinkSearch):
    """Search the web for links using the Google Serper API"""

    _SE_NAME = "Google Serper API"
    _URL = "https://google.serper.dev/search"

    API_KEY_VAR = "SERPER_API_KEY"
    """Environment variable that should contain the Google Serper API key"""

    def __init__(self, api_key=None, verify=False):
        """

        Parameters
        ----------
        api_key : str, optional
            API key for serper search API. If ``None``, will look up the
            API key using the ``"SERPER_API_KEY"`` environment variable.
            By default, ``None``.
        verify : bool, default=False
            Option to use SSL verification when making request to API
            endpoint. By default, ``False``.
        """
        super().__init__(api_key=api_key)
        self.verify = verify

    async def _search(self, query, num_results=10):
        """Search web for links related to a query"""

        payload = json.dumps({"q": query, "num": num_results})
        headers = {'X-API-KEY': self.api_key,
                   'Content-Type': 'application/json'}

        response = requests.request("POST", self._URL, headers=headers,
                                    data=payload, verify=self.verify)
        results = json.loads(response.text).get('organic', {})
        return list(filter(None, (result.get("link", "").replace("+", "%20")
                                  for result in results)))



# -*- coding: utf-8 -*-
# fmt: off
"""
ELM PDF to text parser
"""
import os
import subprocess
import requests
import tempfile
import copy
from PyPDF2 import PdfReader
import logging

from elm.base import ApiBase
from elm.utilities.parse import is_multi_col, combine_pages, clean_headers


logger = logging.getLogger(__name__)


class PDFtoTXT(ApiBase):
    """Class to parse text from a PDF document."""

    MODEL_ROLE = ('You clean up poorly formatted text '
                  'extracted from PDF documents.')
    """High level model role."""

    MODEL_INSTRUCTION = ('Text extracted from a PDF: '
                         '\n"""\n{}\n"""\n\n'
                         'The text above was extracted from a PDF document. '
                         'Can you make it nicely formatted? '
                         'Please only return the formatted text '
                         'without comments or added information.')
    """Instructions to the model with python format braces for pdf text"""

    def __init__(self, fp, page_range=None, model=None):
        """
        Parameters
        ----------
        fp : str
            Filepath to .pdf file to extract.
        page_range : None | list
            Optional 2-entry list/tuple to set starting and ending pages
            (python indexing)
        model : None | str
            Optional specification of OpenAI model to use. Default is
            cls.DEFAULT_MODEL
        """
        super().__init__(model)
        self.fp = fp
        self.raw_pages = self.load_pdf(page_range)
        self.pages = self.raw_pages
        self.full = combine_pages(self.raw_pages)

    def load_pdf(self, page_range):
        """Basic load of pdf to text strings

        Parameters
        ----------
        page_range : None | list
            Optional 2-entry list/tuple to set starting and ending pages
            (python indexing)

        Returns
        -------
        out : list
            List of strings where each entry is a page. This is the raw PDF
            text before GPT cleaning
        """

        logger.info('Loading PDF: {}'.format(self.fp))
        out = []
        pdf = PdfReader(self.fp)

        if page_range is not None:
            assert len(page_range) == 2
            page_range = slice(*page_range)
        else:
            page_range = slice(0, None)

        for i, page in enumerate(pdf.pages[page_range]):
            page_text = page.extract_text()
            if len(page_text.strip()) == 0:
                logger.debug('Skipping empty page {} out of {}'
                             .format(i + 1 + page_range.start, len(pdf.pages)))
            else:
                out.append(page_text)

        logger.info('Finished loading PDF.')
        return out

    def make_gpt_messages(self, pdf_raw_text):
        """Make the chat completion messages list for input to GPT

        Parameters
        ----------
        pdf_raw_text : str
            Raw PDF text to be cleaned

        Returns
        -------
        messages : list
            Messages for OpenAI chat completion model. Typically this looks
            like this:
                [{"role": "system", "content": "You do this..."},
                 {"role": "user", "content": "Please do this: {}"}]
        """

        query = self.MODEL_INSTRUCTION.format(pdf_raw_text)
        messages = [{"role": "system", "content": self.MODEL_ROLE},
                    {"role": "user", "content": query}]

        return messages

    def clean_txt(self):
        """Use GPT to clean raw pdf text in serial calls to the OpenAI API.

        Returns
        -------
        clean_pages : list
            List of clean text strings where each list entry is a page from the
            PDF
        """

        logger.info('Cleaning PDF text...')
        clean_pages = []

        for i, raw_page in enumerate(self.raw_pages):
            msg = self.make_gpt_messages(copy.deepcopy(raw_page))
            req = {"model": self.model, "messages": msg, "temperature": 0.0}

            kwargs = dict(url=self.URL, headers=self.HEADERS, json=req)

            try:
                response = requests.post(**kwargs)
                response = response.json()
            except Exception as e:
                msg = 'Error in OpenAI API call!'
                logger.exception(msg)
                response = {'error': str(e)}

            choice = response.get('choices', [{'message': {'content': ''}}])[0]
            message = choice.get('message', {'content': ''})
            content = message.get('content', '')
            clean_pages.append(content)
            logger.debug('Cleaned page {} out of {}'
                         .format(i + 1, len(self.raw_pages)))

        logger.info('Finished cleaning PDF.')

        self.pages = clean_pages
        self.full = combine_pages(self.pages)
        self.validate_clean()

        return clean_pages

    async def clean_txt_async(self, ignore_error=None, rate_limit=40e3):
        """Use GPT to clean raw pdf text in parallel calls to the OpenAI API.

        NOTE: you need to call this using the await command in ipython or
        jupyter, e.g.: `out = await PDFtoTXT.clean_txt_async()`

        Parameters
        ----------
        ignore_error : None | callable
            Optional callable to parse API error string. If the callable
            returns True, the error will be ignored, the API call will not be
            tried again, and the output will be an empty string.
        rate_limit : float
            OpenAI API rate limit (tokens / minute). Note that the
            gpt-3.5-turbo limit is 90k as of 4/2023, but we're using a large
            factor of safety (~1/2) because we can only count the tokens on the
            input side and assume the output is about the same count.

        Returns
        -------
        clean_pages : list
            List of clean text strings where each list entry is a page from the
            PDF
        """

        logger.info('Cleaning PDF text asyncronously...')

        all_request_jsons = []
        for page in self.raw_pages:
            msg = self.make_gpt_messages(page)
            req = {"model": self.model, "messages": msg, "temperature": 0.0}
            all_request_jsons.append(req)

        clean_pages = await self.call_api_async(self.URL, self.HEADERS,
                                                all_request_jsons,
                                                ignore_error=ignore_error,
                                                rate_limit=rate_limit)

        for i, page in enumerate(clean_pages):
            choice = page.get('choices', [{'message': {'content': ''}}])[0]
            message = choice.get('message', {'content': ''})
            content = message.get('content', '')
            clean_pages[i] = content

        logger.info('Finished cleaning PDF.')

        self.pages = clean_pages
        self.full = combine_pages(self.pages)
        self.validate_clean()

        return clean_pages

    def clean_poppler(self, layout=True):
        """Clean the pdf using the poppler pdftotxt utility

        Requires the `pdftotext` command line utility from this software:
            https://poppler.freedesktop.org/

        Parameters
        ----------
        layout : bool
            Layout flag for poppler pdftotxt utility: "maintain original
            physical layout". Layout=True works well for single column text,
            layout=False collapses the double columns into single columns which
            works better for downstream chunking and LLM work.

        Returns
        -------
        out : str
            Joined cleaned pages
        """

        with tempfile.TemporaryDirectory() as td:
            fp_out = os.path.join(td, 'poppler_out.txt')
            args = ['pdftotext', f"{self.fp}", f"{fp_out}"]
            if layout:
                args.insert(1, '-layout')

            if not os.path.exists(os.path.dirname(fp_out)):
                os.makedirs(os.path.dirname(fp_out), exist_ok=True)

            try:
                stdout = subprocess.run(args, check=True,
                                        stdout=subprocess.PIPE)
            except Exception as e:
                msg = ('PDF cleaning with poppler failed! This usually '
                       'because you have not installed the poppler utility '
                       '(see https://poppler.freedesktop.org/). '
                       f'Full error: {e}')
                logger.exception(msg)
                raise RuntimeError(msg) from e
            else:
                if stdout.returncode != 0:
                    msg = ('Poppler raised return code {}: {}'
                           .format(stdout.returncode, stdout))
                    logger.exception(msg)
                    raise RuntimeError(msg)

            with open(fp_out, 'r') as f:
                clean_txt = f.read()

        # break on poppler page break
        self.pages = clean_txt.split('\x0c')
        remove = []
        for i, page in enumerate(self.pages):
            if not any(page.strip()):
                remove.append(i)
        for i in remove[::-1]:
            _ = self.pages.pop(i)

        self.full = combine_pages(self.pages)

        return self.full

    def validate_clean(self):
        """Run some basic checks on the GPT cleaned text vs. the raw text"""
        repl = ('\n', '.', ',', '-', '/', ':')

        if not any(self.full.replace('\n', '').strip()):
            msg = 'Didnt get ANY clean output text!'
            logger.error(msg)
            raise RuntimeError(msg)

        def replace_chars_for_clean(text):
            for char in repl:
                text = text.replace(char, ' ')
            return text

        for i, (raw, clean) in enumerate(zip(self.raw_pages, self.pages)):
            raw_words = replace_chars_for_clean(raw).split(' ')
            clean_words = replace_chars_for_clean(clean).split(' ')

            raw_words = {x for x in raw_words if len(x) > 2}
            clean_words = {x for x in clean_words if len(x) > 2}

            isin = sum(x in clean_words for x in raw_words)

            perc = 100
            if isin > 0 and len(raw_words) > 0:
                perc = 100 * isin / len(raw_words)

            if perc < 70:
                logger.warning('Page {} of {} has a {:.2f}% match with {} '
                               'unique words in the raw text.'
                               .format(i + 1, len(self.raw_pages), perc,
                                       len(raw_words)))
            else:
                logger.info('Page {} of {} has a {:.2f}% match with {} '
                            'unique words in the raw text.'
                            .format(i + 1, len(self.raw_pages), perc,
                                    len(raw_words)))

    def clean_headers(self, char_thresh=0.6, page_thresh=0.8, split_on='\n',
                      iheaders=(0, 1, -2, -1)):
        """Clean headers/footers that are duplicated across pages

        Parameters
        ----------
        char_thresh : float
            Fraction of characters in a given header that are similar between
            pages to be considered for removal
        page_thresh : float
            Fraction of pages that share the header to be considered for
            removal
        split_on : str
            Chars to split lines of a page on
        iheaders : list | tuple
            Integer indices to look for headers after splitting a page into
            lines based on split_on. This needs to go from the start of the
            page to the end.

        Returns
        -------
        out : str
            Clean text with all pages joined
        """
        self.pages = clean_headers(self.pages, char_thresh=char_thresh,
                                   page_thresh=page_thresh, split_on=split_on,
                                   iheaders=iheaders)
        self.full = combine_pages(self.pages)
        return self.full

    def convert_to_txt(self, txt_fp=None, separator='    ',
                       clean_header_kwargs=None):
        """Function to convert contents of pdf document to txt file using
        poppler.

        Parameters
        ----------
        txt_fp: str, optional
            Optional Directory for output txt file.
        separator : str, optional
            Heuristic split string to look for spaces between columns
        clean_header_kwargs : dict, optional
            Optional kwargs to override clean_headers kwargs

        Returns
        -------
        text : str
            Text string containing contents from pdf
        """
        text = self.clean_poppler(layout=True)
        if is_multi_col(text, separator=separator):
            text = self.clean_poppler(layout=False)

        clean_header_kwargs = clean_header_kwargs or {}
        text = self.clean_headers(**clean_header_kwargs)

        if txt_fp is not None:
            with open(txt_fp, 'w') as f:
                f.write(text)
                logger.info(f'Saved: {txt_fp}')

        return text



# -*- coding: utf-8 -*-
"""ELM Web Scraping - DuxDistributedGlobalSearch"""
import logging

from ddgs import DDGS

from elm.web.search.base import SearchEngineLinkSearch


logger = logging.getLogger(__name__)


class DuxDistributedGlobalSearch(SearchEngineLinkSearch):
    """Search the web for links using DuxDistributedGlobalSearch"""

    _SE_NAME = "DuxDistributedGlobalSearch"

    def __init__(self, region="us-en", safesearch="moderate", timelimit=None,
                 page=1, backend=("google", "bing", "yahoo", "duckduckgo"),
                 timeout=10, verify=False):
        """

        Parameters
        ----------
        region : str, optional
            DuxDistributedGlobalSearch search region param.
            By default, ``"us-en"``.
        safesearch : {on, moderate, off}, optional
            The `safesearch` setting for search engines.
            By default, ``None``.
        timelimit : {d, w, m, y}, optional
            The time limit used to bound the search results:

                -d: last day
                -w: last week
                -m: last month
                -y: last year

            By default, ``None``.
        page : int, default=1
            The page of results to return. By default, ``1``.
        backend : str or iter of str, optional
            Option for DuxDistributedGlobalSearch backend:

                - auto: Randomly select 3 search engines to use
                - all: All available search engines are used
                - wikipedia: Wikipedia
                - google: Google
                - bing: Bing
                - brave: Brave
                - mojeek: Mojeek
                - yahoo: Yahoo
                - yandex: Yandex
                - duckduckgo: Duckduckgo

            Can also be a list or tuple of a combination of these.
            By default, ``("google", "bing", "yahoo", "duckduckgo")``.
        timeout : int, optional
            Timeout for HTTP requests, in seconds. By default, ``10``.
        verify : bool, optional
            Apply SSL verification when making the request.
            By default, ``False``.
        """
        self.region = region
        self.safesearch = safesearch
        self.timelimit = timelimit
        self.page = page
        self.backend = backend
        self.timeout = timeout
        self.verify = verify

    async def _search(self, query, num_results=10):
        """Search web for links related to a query"""

        ddgs = DDGS(timeout=self.timeout, verify=self.verify)
        results = ddgs.text(query, region=self.region,
                            safesearch=self.safesearch,
                            timelimit=self.timelimit,
                            page=self.page,
                            backend=self.backend,
                            num_results=num_results)

        return list(filter(None, (info.get('href', "").replace("+", "%20")
                                  for info in results)))




# -*- coding: utf-8 -*-
"""ELM retry utilities."""
import time
import random
import asyncio
import logging
from functools import wraps

import openai

from elm.exceptions import ELMRuntimeError


logger = logging.getLogger(__name__)


def retry_with_exponential_backoff(
    base_delay=1,
    exponential_base=4,
    jitter=True,
    max_retries=3,
    errors=(openai.RateLimitError, openai.APITimeoutError),
):
    """Retry a synchronous function with exponential backoff.

    This decorator works out-of-the-box for OpenAI chat completions
    calls. To configure it for other functions, set the `errors` input
    accordingly.

    Parameters
    ----------
    base_delay : int, optional
        The base delay time, in seconds. This time will be multiplied by
        the exponential_base (plus any jitter) during each retry
        iteration. The multiplication applies *at the first retry*.
        Therefore, if your base delay is ``1`` and your
        `exponential_base` is ``4`` (with no jitter), the delay before
        the first retry will be ``1 * 4 = 4`` seconds. The subsequent
        delay will be ``4 * 4 = 16`` seconds, and so on.
        By default, ``1``.
    exponential_base : int, optional
        The multiplication factor applied to the base `delay` input.
        See description of `delay` for an example. By default, ``4``.
    jitter : bool, optional
        Option to include a random fractional adder (0 - 1) to the
        `exponential_base` before multiplying by the `delay`. This can
        help ensure each function call is submitted slightly offset from
        other calls in a batch and therefore help avoid repeated rate
        limit failures by a batch of submissions arriving simultaneously
        to a service. By default, ``True``.
    max_retries : int, optional
        Max number of retries before raising an `ELMRuntimeError`.
        By default, ``3``.
    errors : tuple, optional
        The error class(es) to signal a retry. Other errors will be
        propagated without retrying.
        By default, ``(openai.RateLimitError, openai.APITimeoutError)``.

    References
    ----------
    https://github.com/openai/openai-cookbook/blob/main/examples/How_to_handle_rate_limits.ipynb
    https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            num_retries = 0
            delay = base_delay

            while True:
                try:
                    return func(*args, **kwargs)
                except errors as e:
                    num_retries = _handle_retries(num_retries, max_retries, e)
                    delay = _compute_delay(delay, exponential_base, jitter)
                    logger.info(
                        "Error: %s. Retrying in %.2f seconds.", str(e), delay
                    )
                    kwargs = _double_timeout(**kwargs)
                    time.sleep(delay)

        return wrapper

    return decorator


def async_retry_with_exponential_backoff(
    base_delay=1,
    exponential_base=4,
    jitter=True,
    max_retries=3,
    errors=(openai.RateLimitError, openai.APITimeoutError),
):
    """Retry an asynchronous function with exponential backoff.

    This decorator works out-of-the-box for OpenAI chat completions
    calls. To configure it for other functions, set the `errors` input
    accordingly.

    Parameters
    ----------
    base_delay : int, optional
        The base delay time, in seconds. This time will be multiplied by
        the exponential_base (plus any jitter) during each retry
        iteration. The multiplication applies *at the first retry*.
        Therefore, if your base delay is ``1`` and your
        `exponential_base` is ``4`` (with no jitter), the delay before
        the first retry will be ``1 * 4 = 4`` seconds. The subsequent
        delay will be ``4 * 4 = 16`` seconds, and so on.
        By default, ``1``.
    exponential_base : int, optional
        The multiplication factor applied to the base `delay` input.
        See description of `delay` for an example. By default, ``4``.
    jitter : bool, optional
        Option to include a random fractional adder (0 - 1) to the
        `exponential_base` before multiplying by the `delay`. This can
        help ensure each function call is submitted slightly offset from
        other calls in a batch and therefore help avoid repeated rate
        limit failures by a batch of submissions arriving simultaneously
        to a service. By default, ``True``.
    max_retries : int, optional
        Max number of retries before raising an `ELMRuntimeError`.
        By default, ``3``.
    errors : tuple, optional
        The error class(es) to signal a retry. Other errors will be
        propagated without retrying.
        By default, ``(openai.RateLimitError, openai.APITimeoutError)``.

    References
    ----------
    https://github.com/openai/openai-cookbook/blob/main/examples/How_to_handle_rate_limits.ipynb
    https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            num_retries = 0
            delay = base_delay

            while True:
                try:
                    return await func(*args, **kwargs)
                except errors as e:
                    num_retries = _handle_retries(num_retries, max_retries, e)
                    delay = _compute_delay(delay, exponential_base, jitter)
                    logger.info(
                        "Error: %s. Retrying in %.2f seconds.", str(e), delay
                    )
                    kwargs = _double_timeout(**kwargs)
                    await asyncio.sleep(delay)

        return wrapper

    return decorator


def _handle_retries(num_retries, max_retries, error):
    """Raise error if retry attempts exceed max limit"""
    num_retries += 1
    if num_retries > max_retries:
        msg = f"Maximum number of retries ({max_retries}) exceeded"
        raise ELMRuntimeError(msg) from error
    return num_retries


def _compute_delay(delay, exponential_base, jitter):
    """Compute the next delay time"""
    return delay * exponential_base * (1 + jitter * random.random())


def _double_timeout(**kwargs):
    """Double timeout parameter if it exists in kwargs."""
    if "timeout" not in kwargs:
        return kwargs

    prev_timeout = kwargs["timeout"]
    logger.info(
        "Detected 'timeout' key in kwargs. Doubling this input from "
        "%.2f to %.2f for next iteration.",
        prev_timeout,
        prev_timeout * 2,
    )
    kwargs["timeout"] = prev_timeout * 2
    return kwargs
